"""Object to represent the result definition used by the failure operator in DPF Composites."""

from dataclasses import dataclass, field
from enum import Enum
import json
from typing import Any, Dict, List, Optional, Sequence, Union

from ._typing_helper import PATH as _PATH
from .failure_criteria import CombinedFailureCriterion

__all__ = ("FailureMeasureEnum", "ResultDefinitionScope", "ResultDefinition")


class FailureMeasureEnum(str, Enum):
    """Provides available failure measures."""

    # TODO: Add descriptions for each property
    INVERSE_RESERVE_FACTOR: str = "inverse_reserve_factor"
    MARGIN_OF_SAFETY: str = "safety_margin"
    RESERVE_FACTOR: str = "safety_factor"


_SUPPORTED_EXPRESSIONS = ["composite_failure"]
_SUPPORTED_MEASURES = [v.value for v in FailureMeasureEnum]
_SUPPORTED_STRESS_STRAIN_EVAL_MODES = ["rst_file", "mapdl_live"]


@dataclass
class ResultDefinitionScope:
    """Provides the result definition scope."""

    composite_definition: _PATH
    element_scope: Sequence[int] = field(default_factory=lambda: [])
    ply_scope: Sequence[str] = field(default_factory=lambda: [])
    named_selection_scope: Sequence[str] = field(default_factory=lambda: [])
    """Assembly files that define the mapping of the labels.

    This attribute is needed if multiple parts are assembled in Workbench or
    Mechanical to map the local element and node labels to the global labels.
    """
    mapping_file: Optional[_PATH] = None
    """Path to the mapping file for all element labels in the element scope.
    """

    write_data_for_full_element_scope: bool = True
    """Whether to write the data for all element labels in the element scope.

    This makes sense if an element scope is explicitly requested
    but the actual scope where postprocessing has happened is smaller,
    perhaps due to ply scoping.
    """


class ResultDefinition:
    """Represents the result definition of DPF Composites.

    This class is used to configure the DPF operators ``composite::failure_evaluator``
    and ``composite::sampling_point_evaluator``.

    """

    _VERSION = 1
    _ACCUMULATOR = "max"

    def __init__(
        self,
        name: str,
        combined_failure_criterion: CombinedFailureCriterion,
        composite_scopes: Sequence[ResultDefinitionScope],
        rst_files: Sequence[_PATH],
        material_file: _PATH,
        measure: str = "inverse_reserve_factor",
        stress_strain_eval_mode: str = "rst_file",
        time: Union[float, None] = None,
        expression: str = "composite_failure",
        max_chunk_size: int = 50000,
    ):
        """Create a ResultDefinition object."""
        self._name = name
        self._expression = expression
        self._combined_failure_criterion = combined_failure_criterion
        self._measure = measure
        self._composite_scopes = composite_scopes
        self._material_file = material_file
        self._rst_files = list(rst_files)
        self._stress_strain_eval_mode = stress_strain_eval_mode
        self._time = time
        self._max_chunk_size = max_chunk_size

    @property
    def name(self) -> str:
        """Custom name."""
        return self._name

    @name.setter
    def name(self, value: str) -> None:
        self._name = value

    @property
    def expression(self) -> str:
        """Type of the result. The supported type is ``"composite_failure"``."""
        return self._expression

    @expression.setter
    def expression(self, value: str) -> None:
        if value in _SUPPORTED_EXPRESSIONS:
            self._expression = value
        else:
            values = ", ".join([v for v in _SUPPORTED_EXPRESSIONS])
            raise ValueError(
                f"Expression {value} is not allowed. Supported expressions are {values}."
            )

    @property
    def combined_failure_criterion(self) -> CombinedFailureCriterion:
        """Configuration of the failure criteria such as maximum stress, puck, and wrinkling."""
        return self._combined_failure_criterion

    @combined_failure_criterion.setter
    def combined_failure_criterion(self, cfc: CombinedFailureCriterion) -> None:
        self._combined_failure_criterion = cfc

    @property
    def measure(self) -> str:
        """Return type of the failure values.

        Supported types are ``"inverse_reserve_factor"``, ``"safety_factor"``,
        and ``"safety_margin"``.
        """
        return self._measure

    @measure.setter
    def measure(self, value: str) -> None:
        if value not in _SUPPORTED_MEASURES:
            values = ", ".join([v for v in _SUPPORTED_MEASURES])
            raise ValueError(f"Measure {value} is not allowed. Supported measures are {values}.")
        else:
            self._measure = value

    @property
    def scopes(self) -> Sequence[ResultDefinitionScope]:
        """Scopes of the result definition."""
        return self._composite_scopes

    @scopes.setter
    def scopes(self, value: Sequence[ResultDefinitionScope]) -> None:
        self._composite_scopes = value

    @property
    def rst_files(self) -> List[_PATH]:
        """Path of the result (RST) files."""
        return self._rst_files

    @rst_files.setter
    def rst_files(self, value: List[_PATH]) -> None:
        self._rst_files = list(value)

    @property
    def material_file(self) -> _PATH:
        """Path of the material files that store the material properties.

        Supported formats are XML and ENGD.
        """
        return self._material_file

    @material_file.setter
    def material_file(self, value: _PATH) -> None:
        self._material_file = value

    @property
    def stress_strain_eval_mode(self) -> str:
        """Results loaded from a result (RST) file by default.

        You can set this property to ``"mapdl_live"`` to activate on-the-fly
        strain and stress evaluation. This property can be used if the result
        file contains only the primary results (deformations).
        """
        return self._stress_strain_eval_mode

    @stress_strain_eval_mode.setter
    def stress_strain_eval_mode(self, value: str) -> None:
        if value in _SUPPORTED_STRESS_STRAIN_EVAL_MODES:
            self._stress_strain_eval_mode = value
        else:
            values = ", ".join([v for v in _SUPPORTED_STRESS_STRAIN_EVAL_MODES])
            raise ValueError(
                f"Stress strain evaluation mode '{value}' is not allowed. "
                f"Supported values are {values}."
            )

    @property
    def time(self) -> Union[float, None]:
        """Time or solution step.

        DPF Composites automatically selects the last time step if time is not set.

        You can use the :meth:`.CompositeModel.get_result_times_or_frequencies` method
        to list the available times or frequencies in the result file.
        """
        return self._time

    @time.setter
    def time(self, value: Union[float, None]) -> None:
        self._time = value

    @property
    def max_chunk_size(self) -> int:
        """Maximum chunk size (number of elements) for the result evaluation.

        Small chunks reduce the maximum peak of memory, but too many chunks causes
        some overhead. The default is 50,000. Use ``-1`` to disable chunking.
        """
        return self._max_chunk_size

    @max_chunk_size.setter
    def max_chunk_size(self, value: int) -> None:
        self._max_chunk_size = value

    def to_dict(self) -> Dict[str, Any]:
        """Get the result definition in a dictionary representation."""
        cfc = self.combined_failure_criterion
        if not cfc:
            raise ValueError("Combined failure criterion is not defined.")

        if self.measure not in _SUPPORTED_MEASURES:
            values = ", ".join([v for v in _SUPPORTED_MEASURES])
            raise ValueError(
                f"Measure `{self.measure}` is invalid. Supported measures are {values}."
            )

        result_definition = {
            "version": self._VERSION,
            "accumulator": "max",
            "expression": f"{self.expression}",
            "failure_criteria_definition": {cfc.JSON_DICT_KEY: cfc.to_dict()},
            "measures": [self.measure],
            "stress_strain_eval_mode": f"{self.stress_strain_eval_mode}",
            "max_chunk_size": self.max_chunk_size,
        }

        if self.time:
            result_definition["time"] = self.time

        def get_scope(
            result_definition_scope: ResultDefinitionScope,
            rst_files: List[_PATH],
            material_file: _PATH,
        ) -> Dict[str, Any]:
            write_for_full_scope = result_definition_scope.write_data_for_full_element_scope
            mapping_entry = []
            if result_definition_scope.mapping_file is not None:
                mapping_entry.append(str(result_definition_scope.mapping_file))
            return {
                "datasources": {
                    "composite_definition": [str(result_definition_scope.composite_definition)],
                    "assembly_mapping_file": mapping_entry,
                    "rst_file": [str(filename) for filename in rst_files],
                    "material_file": [str(material_file)],
                },
                "write_data_for_full_element_scope": write_for_full_scope,
                "elements": [int(v) for v in result_definition_scope.element_scope],
                "ply_ids": result_definition_scope.ply_scope,
                "named_selections": result_definition_scope.named_selection_scope,
            }

        scopes = {
            "scopes": [
                get_scope(scope, self.rst_files, self.material_file)
                for scope in self._composite_scopes
            ]
        }

        result_definition.update(scopes)
        return result_definition

    def to_json(self) -> str:
        """Convert the dictionary representation of the result definition to a JSON dictionary."""
        return json.dumps(self.to_dict())

    def _get_properties(self, exclude: Sequence[str] = tuple()) -> Sequence[Any]:
        properties = [
            attr
            for attr in dir(self)
            if not attr.startswith("__")
            and not attr.startswith("_")
            and not callable(getattr(self, attr))
            and attr not in exclude
        ]

        return properties

    def check_has_single_scope(self, msg: str) -> None:
        """Check that the result definition has one scope."""
        if len(self.scopes) != 1:
            raise RuntimeError(f"Result definition has multiple scopes. {msg}")

    def _short_descr(self) -> str:
        """:return: short description of the object."""
        return f"{self.__class__.__name__}(name='{self.name}')"

    def __repr__(self) -> str:
        """:return: string conversion."""
        s_attrs = ", ".join([f"{attr}={getattr(self, attr)}" for attr in self._get_properties()])
        s = f"{self.__class__.__name__}({s_attrs})"
        return s



"""Lay-up information provider."""

from dataclasses import dataclass
from typing import Any, Collection, Dict, List, Optional, Sequence, Union, cast

import ansys.dpf.core as dpf
from ansys.dpf.core import DataSources, MeshedRegion, Operator, PropertyField
from ansys.dpf.core.server_types import BaseServer
import numpy as np
from numpy.typing import NDArray

from .._indexer import (
    FieldIndexerNoDataPointer,
    FieldIndexerWithDataPointer,
    PropertyFieldIndexerArrayValue,
    PropertyFieldIndexerNoDataPointer,
    PropertyFieldIndexerNoDataPointerNoBoundsCheck,
    PropertyFieldIndexerSingleValue,
    PropertyFieldIndexerWithDataPointer,
    PropertyFieldIndexerWithDataPointerNoBoundsCheck,
)
from ..server_helpers import version_older_than
from ._enums import LayupProperty


def _get_separator(server: BaseServer) -> str:
    if version_older_than(server, "7.0"):
        return ":"
    else:
        return "::"


def _get_analysis_ply_prefix(server: BaseServer) -> str:
    return "AnalysisPly" + _get_separator(server)


def get_all_analysis_ply_names(mesh: MeshedRegion) -> Collection[str]:
    """Get names of all available analysis plies."""
    prefix = _get_analysis_ply_prefix(mesh._server)  # pylint: disable=protected-access
    return [
        property_field_name[len(prefix) :]
        for property_field_name in mesh.available_property_fields
        if property_field_name.startswith(prefix)
    ]


def _get_analysis_ply(mesh: MeshedRegion, name: str, skip_check: bool = False) -> PropertyField:
    prefix = _get_analysis_ply_prefix(mesh._server)  # pylint: disable=protected-access
    property_field_name = prefix + name

    # Because this test can be slow, it can be skipped
    if not skip_check and property_field_name not in mesh.available_property_fields:
        available_analysis_plies = get_all_analysis_ply_names(mesh)
        raise RuntimeError(
            f"Analysis ply is not available: {name}. "
            f"Available analysis plies: {available_analysis_plies}"
        )
    return mesh.property_field(property_field_name)


@dataclass(frozen=True)
class ElementInfo:
    """Provides lay-up information for an element.

    Use the :class:`~ElementInfoProvider` class to obtain the
    :class:`~ElementInfo` class for an element.

    Parameters
    ----------
    id
        Element ID or label.
    n_layers
        Number of layers. For non-layered elements, the value is ``1``.
    n_corner_nodes
        Number of corner nodes (without midside nodes).
    n_spots
        Number of spots (through-the-thickness integration points) per layer.
    element_type
        APDL element type. For example, ``181`` for layered shells.
    dpf_material_ids
        List of DPF material IDs for all layers.
    is_shell
        Whether the element is a shell element.
    number_of_nodes_per_spot_plane
        Number of nodes per output plane. The value is equal
        to ``n_corner_nodes`` for shell elements and ``n_corner_nodes``
        divided by two for layered solid elements. The value is equal to ``-1``
        for non-layered elements.
    """

    id: int
    n_layers: int
    n_corner_nodes: int
    n_spots: int
    is_layered: bool
    element_type: int
    dpf_material_ids: NDArray[np.int64]
    is_shell: bool
    number_of_nodes_per_spot_plane: int


_supported_element_types = [181, 281, 185, 186, 187, 190]

"""
Map of keyopt_8 to number of spots.
Example: Element 181 with keyopt8==1 has two spots
"""
_n_spots_by_element_type_and_keyopt_dict: Dict[int, Dict[int, int]] = {
    181: {0: 0, 1: 2, 2: 3},
    281: {0: 0, 1: 2, 2: 3},
    185: {0: 0, 1: 2},
    186: {0: 0, 1: 2},
    187: {0: 0},
    190: {0: 0, 1: 2},
}


def _is_shell(apdl_element_type: np.int64) -> bool:
    return {181: True, 281: True, 185: False, 186: False, 187: False, 190: False}[
        int(apdl_element_type)
    ]


def _get_n_spots(apdl_element_type: np.int64, keyopt_8: np.int64, keyopt_3: np.int64) -> int:
    if keyopt_3 == 0:
        if apdl_element_type == 185 or apdl_element_type == 186:
            return 0

    try:
        return _n_spots_by_element_type_and_keyopt_dict[int(apdl_element_type)][int(keyopt_8)]
    except KeyError as exc:
        raise RuntimeError(
            f"Unsupported element type keyopt8 combination "
            f"Apdl Element Type: {apdl_element_type} "
            f"keyopt8: {keyopt_8}."
        ) from exc


def _get_corner_nodes_by_element_type_array() -> NDArray[np.int64]:
    # Precompute n_corner_nodes for all element types
    # corner_nodes_by_element_type by can be indexed by element type to get the number of
    # corner nodes. If negative value is returned number of corner nodes is not available.
    all_element_types = [int(e.value) for e in dpf.element_types if e.value >= 0]
    corner_nodes_by_element_type: NDArray[np.int64] = np.full(
        np.amax(all_element_types) + 1, -1, dtype=np.int64
    )

    corner_nodes_by_element_type[all_element_types] = [
        dpf.element_types.descriptor(element_type).n_corner_nodes
        if dpf.element_types.descriptor(element_type).n_corner_nodes is not None
        else -1
        for element_type in all_element_types
    ]
    return corner_nodes_by_element_type


class AnalysisPlyInfoProvider:
    """AnalysisPlyInfoProvider. Can be used to compute the layer indices of a given analysis ply.

    Parameters
    ----------
    mesh
        DPF MeshedRegion with lay-up information.
    name
        Analysis Ply Name
    """

    def __init__(self, mesh: MeshedRegion, name: str):
        """Initialize AnalysisPlyProvider."""
        self.name = name
        self.property_field = _get_analysis_ply(mesh, name)
        self._layer_indices = PropertyFieldIndexerNoDataPointer(self.property_field)

    def get_layer_index_by_element_id(self, element_id: int) -> Optional[np.int64]:
        """Get the layer index for the analysis ply in a given element.

        Parameters
        ----------
        element_id:
            Element Id/Label

        """
        return self._layer_indices.by_id(element_id)

    def ply_element_ids(self) -> Sequence[np.int64]:
        """Return list of element labels of the analysis ply."""
        return cast(Sequence[np.int64], self.property_field.scoping.ids)


def get_dpf_material_id_by_analyis_ply_map(
    mesh: MeshedRegion,
    data_source_or_streams_provider: Union[DataSources, Operator],
) -> Dict[str, np.int64]:
    """Get Dict that maps analysis ply names to dpf_material_ids.

    Parameters
    ----------
    mesh
        DPF Meshed region enriched with lay-up information
    data_source_or_streams_provider:
        DPF data source with rst file or streams_provider. The streams provider is
        available from :attr:`.CompositeModel.core_model` (under metadata.streams_provider).
    """
    # Note: The stream_provider_or_data_source is not strictly needed for this workflow
    # We just need it because get_element_info_provider provider needs it (which needs
    # it to determine the keyopts, which are not needed in this context)
    # Maybe we could split the ElementInfoProvider
    analysis_ply_to_material_map = {}
    element_info_provider = get_element_info_provider(
        mesh=mesh, stream_provider_or_data_source=data_source_or_streams_provider
    )
    all_element_ids = mesh.elements.scoping.ids

    for analysis_ply_name in get_all_analysis_ply_names(mesh):
        analysis_ply_info_provider = AnalysisPlyInfoProvider(mesh, analysis_ply_name)

        # Note we check if there is any valid elements in this analysis ply
        # We won't find any elements if all the elements of a ply
        # were suppressed and therefore the analysis ply is not part
        # of the mesh anymore. In this case we simply don't
        # add the ply to the map
        diff = set(analysis_ply_info_provider.ply_element_ids()).intersection(all_element_ids)
        for element_id in diff:
            element_id_int = int(element_id)
            element_info = element_info_provider.get_element_info(element_id_int)
            if element_info is not None:
                layer_index = analysis_ply_info_provider.get_layer_index_by_element_id(
                    element_id_int
                )
                assert (
                    layer_index is not None
                ), f"No layer index found for element with id {element_id_int}."
                analysis_ply_to_material_map[analysis_ply_name] = element_info.dpf_material_ids[
                    layer_index
                ]
                break

    return analysis_ply_to_material_map


def get_analysis_ply_index_to_name_map(
    mesh: MeshedRegion,
) -> Dict[int, str]:
    """Get Dict that maps analysis ply indices to analysis ply names.

    The resulting dict can be used to map from the indices stored in
    mesh.property_field("layer_to_analysis_ply") to the analysis ply name

    Parameters
    ----------
    mesh
        DPF Meshed region enriched with lay-up information
    """
    analysis_ply_name_to_index_map = {}
    with mesh.property_field("layer_to_analysis_ply").as_local_field() as local_field:
        for analysis_ply_name in get_all_analysis_ply_names(mesh):
            analysis_ply_property_field = _get_analysis_ply(
                mesh, analysis_ply_name, skip_check=True
            )
            first_element_id = analysis_ply_property_field.scoping.id(0)
            analysis_ply_indices: List[int] = local_field.get_entity_data_by_id(first_element_id)

            layer_index = analysis_ply_property_field.get_entity_data(0)[0]
            assert layer_index is not None
            analysis_ply_name_to_index_map[analysis_ply_indices[layer_index]] = analysis_ply_name

    return analysis_ply_name_to_index_map


class ElementInfoProvider:
    """Provider for :class:`~ElementInfo`.

    Use :func:`~get_element_info_provider` to create :class:`~ElementInfoProvider`
    objects.

    Initialize the class before a loop and
    call :func:`~get_element_info` for each element.

    Note that the :class:`~ElementInfoProvider` class is not fully supported for
    distributed RST files. The :func:`~get_element_info` method will raise an
    exception if the DPF server version does not support reading the required
    information.

    Parameters
    ----------
    mesh
    layer_indices
    element_types_apdl
    element_types_dpf
    keyopt_8_values
    keyopt_3_values
    material_ids
    no_bounds_checks
        Disable bounds checks.
        Results in better performance but potentially cryptic
        error messages
    """

    def __init__(
        self,
        mesh: MeshedRegion,
        layer_indices: PropertyField,
        element_types_apdl: PropertyField,
        element_types_dpf: PropertyField,
        keyopt_8_values: PropertyField,
        keyopt_3_values: PropertyField,
        material_ids: PropertyField,
        no_bounds_checks: bool = False,
    ):
        """Initialize ElementInfoProvider."""
        # Note: Every property we add to element info adds some performance
        # overhead for all the calls to get_element info. We should keep it
        # focused on the most important properties. We can add different providers
        # for other properties (such as thickness and angles)

        def get_indexer_with_data_pointer(field: PropertyField) -> PropertyFieldIndexerArrayValue:
            if no_bounds_checks:
                return PropertyFieldIndexerWithDataPointerNoBoundsCheck(field)
            else:
                return PropertyFieldIndexerWithDataPointer(field)

        def get_indexer_no_data_pointer(field: PropertyField) -> PropertyFieldIndexerSingleValue:
            if no_bounds_checks:
                return PropertyFieldIndexerNoDataPointerNoBoundsCheck(field)
            else:
                return PropertyFieldIndexerNoDataPointer(field)

        # Has to be always with bounds checks because it does not contain
        # data for all the elements
        self.layer_indices = PropertyFieldIndexerWithDataPointer(layer_indices)
        self.layer_materials = get_indexer_with_data_pointer(material_ids)

        self.apdl_element_types = get_indexer_no_data_pointer(element_types_apdl)
        self.dpf_element_types = get_indexer_no_data_pointer(element_types_dpf)
        self.keyopt_8_values = get_indexer_no_data_pointer(keyopt_8_values)
        self.keyopt_3_values = get_indexer_no_data_pointer(keyopt_3_values)

        self.mesh = mesh
        self.corner_nodes_by_element_type = _get_corner_nodes_by_element_type_array()

    def get_element_info(self, element_id: int) -> Optional[ElementInfo]:
        """Get :class:`~ElementInfo` for a given element id.

        Parameters
        ----------
        element_id:
            Element Id/Label

        Returns
        -------
        Optional[ElementInfo]:
            Returns None if element type is not supported
        """
        is_layered = False
        n_layers = 1

        keyopt_8 = self.keyopt_8_values.by_id(element_id)
        keyopt_3 = self.keyopt_3_values.by_id(element_id)
        apdl_element_type = self.apdl_element_types.by_id(element_id)

        if keyopt_3 is None or keyopt_8 is None or apdl_element_type is None:
            raise RuntimeError(
                "Could not determine element properties. Probably they were requested for an"
                f" invalid element id. Element id: {element_id}\n"
                "Note that creating ElementInfo is not fully supported for distributed RST files."
            )

        if int(apdl_element_type) not in _supported_element_types:
            return None

        n_spots = _get_n_spots(apdl_element_type, keyopt_8, keyopt_3)
        dpf_material_ids: Any = []
        element_type = self.dpf_element_types.by_id(element_id)
        if element_type is None:
            raise IndexError(f"No DPF element type for element with id {element_id}.")

        layer_data = self.layer_indices.by_id(element_id)
        if layer_data is not None:
            dpf_material_ids = self.layer_materials.by_id(element_id)
            assert dpf_material_ids is not None
            assert layer_data[0] + 1 == len(layer_data), "Invalid size of layer data"
            n_layers = layer_data[0]
            is_layered = True

        corner_nodes_dpf = self.corner_nodes_by_element_type[element_type]
        if corner_nodes_dpf < 0:
            raise ValueError(f"Invalid number of corner nodes for element with type {element_type}")
        is_shell = _is_shell(apdl_element_type)
        number_of_nodes_per_spot_plane = -1
        if is_layered:
            number_of_nodes_per_spot_plane = corner_nodes_dpf if is_shell else corner_nodes_dpf // 2
        return ElementInfo(
            n_layers=n_layers,
            n_corner_nodes=corner_nodes_dpf,
            n_spots=n_spots,
            is_layered=is_layered,
            element_type=int(apdl_element_type),
            dpf_material_ids=dpf_material_ids,
            id=element_id,
            is_shell=is_shell,
            number_of_nodes_per_spot_plane=number_of_nodes_per_spot_plane,
        )


def get_element_info_provider(
    mesh: MeshedRegion,
    stream_provider_or_data_source: Union[Operator, DataSources],
    no_bounds_checks: bool = False,
) -> ElementInfoProvider:
    """Get :class:`~ElementInfoProvider` Object.

    Parameters
    ----------
    mesh
    stream_provider_or_data_source
        dpf stream provider or dpf data source
    no_bounds_checks
        Disable bounds checks. Improves
        performance but can result in cryptic error messages

    Returns
    -------
    ElementInfoProvider

    Notes
    -----
        Either a data_source or a stream_provider is required
    """

    def get_keyopt_property_field(keyopt: int) -> PropertyField:
        keyopt_provider = dpf.Operator("mesh_property_provider")
        if isinstance(stream_provider_or_data_source, Operator):
            keyopt_provider.inputs.streams_container(stream_provider_or_data_source)
        else:
            keyopt_provider.inputs.data_sources(stream_provider_or_data_source)

        keyopt_provider.inputs.property_name(f"keyopt_{keyopt}")
        return keyopt_provider.outputs.property_as_property_field()

    requested_property_fields = [
        "apdl_element_type",
        "element_layer_indices",
        "element_layered_material_ids",
    ]

    for property_field_name in requested_property_fields:
        if property_field_name not in mesh.available_property_fields:
            message = f"Missing property field in mesh: '{property_field_name}'."
            if property_field_name in ["element_layer_indices", "element_layer_material_ids"]:
                message += (
                    " Maybe you have to run the lay-up provider operator first. "
                    "Please call add_layup_info_to_mesh "
                )
            raise RuntimeError(message)

    fields = {
        "layer_indices": mesh.property_field("element_layer_indices"),
        "element_types_apdl": mesh.property_field("apdl_element_type"),
        "element_types_dpf": mesh.elements.element_types_field,
        "keyopt_8_values": get_keyopt_property_field(8),
        "keyopt_3_values": get_keyopt_property_field(3),
        "material_ids": mesh.property_field("element_layered_material_ids"),
    }

    return ElementInfoProvider(mesh, **fields, no_bounds_checks=no_bounds_checks)


class LayupPropertiesProvider:
    """Provider for lay-up properties.

    Some properties such as layered dpf_material_ids and
    information about the element type are available
    through the :class:`~ElementInfoProvider`.

    Parameters
    ----------
    layup_provider:
        Use :func:`~add_layup_info_to_mesh` to obtain a lay-up provider.
    mesh
    """

    def __init__(self, layup_provider: Operator, mesh: MeshedRegion):
        """Initialize LayupProperties provider."""
        layup_outputs_container = layup_provider.outputs.section_data_container()
        composite_label = layup_outputs_container.labels[0]
        angle_field = layup_outputs_container.get_field({composite_label: LayupProperty.ANGLE})
        self._angle_indexer = FieldIndexerWithDataPointer(angle_field)
        thickness_field = layup_outputs_container.get_field(
            {composite_label: LayupProperty.THICKNESS}
        )
        self._thickness_indexer = FieldIndexerWithDataPointer(thickness_field)
        shear_angle_field = layup_outputs_container.get_field(
            {composite_label: LayupProperty.SHEAR_ANGLE}
        )
        self._shear_angle_indexer = FieldIndexerWithDataPointer(shear_angle_field)
        offset_field = layup_outputs_container.get_field(
            {composite_label: LayupProperty.LAMINATE_OFFSET}
        )
        self._offset_indexer = FieldIndexerNoDataPointer(offset_field)

        self._index_to_name_map = get_analysis_ply_index_to_name_map(mesh)

        self._analysis_ply_indexer = PropertyFieldIndexerWithDataPointer(
            mesh.property_field("layer_to_analysis_ply")
        )

    def get_layer_angles(self, element_id: int) -> Optional[NDArray[np.double]]:
        """Get angles for all layers. Returns None if element is not layered.

        Parameters
        ----------
        element_id:
            Element Id/Label
        """
        return self._angle_indexer.by_id(element_id)

    def get_layer_thicknesses(self, element_id: int) -> Optional[NDArray[np.double]]:
        """Get thicknesses for all layers. Returns None if element is not layered.

        Parameters
        ----------
        element_id:
            Element Id/Label

        """
        return self._thickness_indexer.by_id(element_id)

    def get_layer_shear_angles(self, element_id: int) -> Optional[NDArray[np.double]]:
        """Get shear angle for all layers. Returns None if element is not layered.

        Parameters
        ----------
        element_id:
            Element Id/Label
        """
        return self._shear_angle_indexer.by_id(element_id)

    def get_element_laminate_offset(self, element_id: int) -> Optional[np.double]:
        """Get laminate offset of element. Returns None if element is not layered.

        Parameters
        ----------
        element_id:
            Element Id/Label

        """
        return self._offset_indexer.by_id(element_id)

    def get_analysis_plies(self, element_id: int) -> Optional[Sequence[str]]:
        """Get analysis ply names. Returns None if element is not layered.

        Parameters
        ----------
        element_id:
            Element Id/Label

        """
        indexes = self._analysis_ply_indexer.by_id(element_id)
        if indexes is None:
            return None
        return [self._index_to_name_map[index] for index in indexes]



"""Hashin failure criterion."""
import inspect

from ._failure_criterion_base import FailureCriterionBase

_DOC_HF = "Whether to activate the failure evaluation regarding fiber failure."
_DOC_HM = "Whether to activate the failure evaluation regarding matrix failure."
_DOC_HD = "Whether to activate the failure evaluation regarding delamination if dim is equal to 3."
_DOC_DIM = (
    "Whether the 2D or 3D formulation of the criterion is used. The latter one also "
    "supports the failure mode delamination."
)
_DOC_WF_HF = "Weighting factor of the fiber failure (hf) mode."
_DOC_WF_HM = "Weighting factor of the matrix failure (hm) mode."
_DOC_WF_HD = "Weighting factor of the delamination (hd) mode."


class HashinCriterion(FailureCriterionBase):
    """Hashin Criterion."""

    __doc__ = f"""Defines the Hashin failure criterion for UD reinforced materials.

    Parameters
    ----------
    hf :
        {_DOC_HF}
    hm :
        {_DOC_HM}
    hd :
        {_DOC_HD}
    dim :
        {_DOC_DIM}
    wf_hf :
        {_DOC_WF_HF}
    wf_hm :
        {_DOC_WF_HM}
    wf_hd :
        {_DOC_WF_HD}

    """

    def __init__(
        self,
        *,
        hf: bool = True,
        hm: bool = True,
        hd: bool = False,
        dim: int = 2,
        wf_hf: float = 1.0,
        wf_hm: float = 1.0,
        wf_hd: float = 1.0,
    ):
        """Create a Hashin failure criterion.

        Can be used in combination with uni-directional orthotropic reinforced materials.
        """
        super().__init__(name="Hashin", active=True)

        for attr in ATTRS_HASHIN:
            setattr(self, attr, eval(attr))

    def _get_hf(self) -> bool:
        return self._hf

    def _set_hf(self, value: bool) -> None:
        self._hf = value

    def _get_hm(self) -> bool:
        return self._hm

    def _set_hm(self, value: bool) -> None:
        self._hm = value

    def _get_hd(self) -> bool:
        return self._hd

    def _set_hd(self, value: bool) -> None:
        self._hd = value

    def _get_dim(self) -> int:
        return self._dim

    def _set_dim(self, value: int) -> None:
        if value in [2, 3]:
            self._dim = value
        else:
            raise AttributeError(
                f"Dimension of {self.name} cannot be set to {value}. Allowed are 2 or 3."
            )

    def _get_wf_hf(self) -> float:
        return self._wf_hf

    def _set_wf_hf(self, value: float) -> None:
        self._wf_hf = value

    def _get_wf_hm(self) -> float:
        return self._wf_hm

    def _set_wf_hm(self, value: float) -> None:
        self._wf_hm = value

    def _get_wf_hd(self) -> float:
        return self._wf_hd

    def _set_wf_hd(self, value: float) -> None:
        self._wf_hd = value

    hf = property(_get_hf, _set_hf, doc=_DOC_HF)
    hm = property(_get_hm, _set_hm, doc=_DOC_HM)
    hd = property(
        _get_hd,
        _set_hd,
        doc=_DOC_HD,
    )
    dim = property(
        _get_dim,
        _set_dim,
        doc=_DOC_DIM,
    )
    wf_hf = property(_get_wf_hf, _set_wf_hf, doc=_DOC_HF)
    wf_hm = property(_get_wf_hm, _set_wf_hm, doc=_DOC_HM)
    wf_hd = property(_get_wf_hd, _set_wf_hd, doc=_DOC_HD)


ATTRS_HASHIN = inspect.signature(HashinCriterion).parameters.keys()



"""Collection of constants used across PyDPF Composites."""

from enum import IntEnum

__all__ = ("Spot", "Sym3x3TensorComponent", "FailureOutput")


class Spot(IntEnum):
    """Implements an interface to access the spots of the results of layered elements."""

    BOTTOM = 1
    MIDDLE = 2
    TOP = 3


class Sym3x3TensorComponent(IntEnum):
    """Provides tensor indices for symmetrical 3x3 tensors."""

    TENSOR11 = 0
    TENSOR22 = 1
    TENSOR33 = 2
    TENSOR21 = 3
    TENSOR31 = 5
    TENSOR32 = 4


class FailureOutput(IntEnum):
    """Provides failure output types.

    The enum value corresponds to the index in the fields container.
    """

    FAILURE_MODE = 0
    FAILURE_VALUE = 1
    MAX_LAYER_INDEX = 2



"""Face Sheet Wrinkling Criterion for sandwich structures."""
import inspect

from ._failure_criterion_base import FailureCriterionBase

_DOC_HOMOGENEOUS_CORE_COEFF = (
    "Wrinkling coefficient (reduction factor) for homogeneous core materials."
)
_DOC_HONEYCOMB_CORE_COEFF = (
    "Wrinkling coefficient (reduction factor) for honeycombs. The default is ``0.33``."
)
_DOC_WF = "Weighting factor of the failure mode (wb or wt). The default is ``0.5``."


class FaceSheetWrinklingCriterion(FailureCriterionBase):
    """Face Sheet Wrinkling Criterion."""

    __doc__ = f"""Defines the face sheet wrinkling criterion for sandwiches (laminate with cores).

    Parameters
    ----------
    homogeneous_core_coeff:
        {_DOC_HOMOGENEOUS_CORE_COEFF}
    honeycomb_core_coeff:
        {_DOC_HONEYCOMB_CORE_COEFF}
    wf:
        {_DOC_WF}
    """

    def __init__(
        self,
        *,
        homogeneous_core_coeff: float = 0.5,
        honeycomb_core_coeff: float = 0.33,
        wf: float = 1.0,
    ):
        """Construct a Face Sheet Wrinkling failure criterion."""
        super().__init__(name="Face Sheet Wrinkling", active=True)

        for attr in ATTRS_WRINKLING:
            setattr(self, attr, eval(attr))

    def _get_homogeneous_core_coeff(self) -> float:
        return self._homogeneous_core_coeff

    def _set_homogeneous_core_coeff(self, value: float) -> None:
        if value > 0:
            self._homogeneous_core_coeff = value
        else:
            raise ValueError("Homogeneous core coefficient must be greater than 0.")

    def _get_honeycomb_core_coeff(self) -> float:
        return self._honeycomb_core_coeff

    def _set_honeycomb_core_coeff(self, value: float) -> None:
        if value > 0:
            self._honeycomb_core_coeff = value
        else:
            raise ValueError("Honeycomb core coefficient must be greater than 0.")

    def _get_wf(self) -> float:
        return self._wf

    def _set_wf(self, value: float) -> None:
        self._wf = value

    homogeneous_core_coeff = property(
        _get_homogeneous_core_coeff,
        _set_homogeneous_core_coeff,
        doc=_DOC_HOMOGENEOUS_CORE_COEFF,
    )
    honeycomb_core_coeff = property(
        _get_honeycomb_core_coeff,
        _set_honeycomb_core_coeff,
        doc=_DOC_HONEYCOMB_CORE_COEFF,
    )

    wf = property(_get_wf, _set_wf, doc=_DOC_WF)


ATTRS_WRINKLING = inspect.signature(FaceSheetWrinklingCriterion).parameters.keys()



"""Composite Model Interface."""
# New interface after 2023 R2
from typing import Any, Callable, Collection, Dict, List, Optional, Sequence, cast
from warnings import warn

import ansys.dpf.core as dpf
from ansys.dpf.core import FieldsContainer, MeshedRegion, Operator, UnitSystem
from ansys.dpf.core.server_types import BaseServer
import numpy as np
from numpy.typing import NDArray

from .composite_scope import CompositeScope
from .data_sources import (
    CompositeDataSources,
    ContinuousFiberCompositesFiles,
    get_composites_data_sources,
)
from .failure_criteria import CombinedFailureCriterion
from .layup_info import (
    ElementInfo,
    LayerProperty,
    LayupPropertiesProvider,
    add_layup_info_to_mesh,
    get_element_info_provider,
)
from .layup_info.material_operators import MaterialOperators, get_material_operators
from .layup_info.material_properties import MaterialProperty, get_constant_property_dict
from .result_definition import FailureMeasureEnum
from .sampling_point import SamplingPointNew
from .server_helpers import upload_continuous_fiber_composite_files_to_server
from .unit_system import get_unit_system


def _deprecated_composite_definition_label(func: Callable[..., Any]) -> Any:
    """Emit a warning when the deprecated ``composite_definition_label`` is used."""
    function_arg = "composite_definition_label"

    def inner(*args: Sequence[Any], **kwargs: Sequence[Any]) -> Any:
        if function_arg in kwargs.keys():
            if kwargs[function_arg]:
                warn(
                    f"Use of {function_arg} is deprecated. Function {func.__name__}."
                    " can be called without this argument.",
                    category=DeprecationWarning,
                    stacklevel=2,
                )
        return func(*args, **kwargs)

    return inner


class CompositeModelImpl:
    """Provides access to the basic composite postprocessing functionality.

    This class supports DPF Server version 7.0 (2024 R1) and later.
    On initialization, the ``CompositeModel`` class automatically adds composite lay-up
    information to the meshed regions. It prepares the providers for different lay-up properties
    so that they can be efficiently evaluated. The composite files provided are automatically
    uploaded to the server if needed.

    .. note::

        When creating a ``CompositeModel`` instance, several providers are created and
        lay-up information is added to the DPF meshed regions. Depending on the use
        case, it can be more efficient to create the providers separately.

    Parameters
    ----------
    composite_files:
        Use the :func:`.get_composite_files_from_workbench_result_folder` function to obtain
        the :class:`.ContinuousFiberCompositesFiles` object.
    server:
        DPF Server to create the model on.
    default_unit_system:
        Unit system that is used if the result file
        does not specify the unit system. This happens
        for pure MAPDL projects.
    """

    def __init__(
        self,
        composite_files: ContinuousFiberCompositesFiles,
        server: BaseServer,
        default_unit_system: Optional[UnitSystem] = None,
    ):
        """Initialize data providers and add composite information to meshed region."""
        self._composite_files = upload_continuous_fiber_composite_files_to_server(
            composite_files, server
        )
        self._data_sources = get_composites_data_sources(self._composite_files)

        self._core_model = dpf.Model(self._data_sources.rst, server=server)
        self._server = server

        self._unit_system = get_unit_system(self._data_sources.rst, default_unit_system)

        self._material_operators = get_material_operators(
            rst_data_source=self._data_sources.material_support,
            unit_system=self._unit_system,
            engineering_data_source=self._data_sources.engineering_data,
        )

        self._layup_provider = add_layup_info_to_mesh(
            mesh=self.get_mesh(),
            data_sources=self.data_sources,
            material_operators=self.material_operators,
            unit_system=self._unit_system,
        )

        self._element_info_provider = get_element_info_provider(
            mesh=self.get_mesh(),
            stream_provider_or_data_source=self._get_rst_streams_provider(),
        )
        self._layup_properties_provider = LayupPropertiesProvider(
            layup_provider=self._layup_provider, mesh=self.get_mesh()
        )

    @property
    def composite_definition_labels(self) -> Sequence[str]:
        """All composite definition labels in the model.

        This property is only relevant for assemblies to access
        plies by ID.
        """
        return list(self.composite_files.composite.keys())

    @property
    def composite_files(self) -> ContinuousFiberCompositesFiles:
        """Composite file paths on the server."""
        return self._composite_files

    @property
    def data_sources(self) -> CompositeDataSources:
        """Composite data sources."""
        return self._data_sources

    @property
    def core_model(self) -> dpf.Model:
        """Underlying DPF core model."""
        return self._core_model

    @property
    def material_operators(self) -> MaterialOperators:
        """Material operators."""
        return self._material_operators

    @_deprecated_composite_definition_label
    def get_mesh(self, composite_definition_label: Optional[str] = None) -> MeshedRegion:
        """Get the underlying DPF meshed region.

        The meshed region contains the lay-up information.
        """
        return self._core_model.metadata.meshed_region

    @_deprecated_composite_definition_label
    def get_layup_operator(self, composite_definition_label: Optional[str] = None) -> Operator:
        """Get the lay-up operator.

        Parameters
        ----------
        composite_definition_label :
            Label of the composite definition, which is the
            dictionary key in the :attr:`.ContinuousFiberCompositesFiles.composite`
            attribute. This parameter is only required for assemblies.
            See the note about assemblies in the description for the :class:`CompositeModel` class.

        """
        return self._layup_provider

    @_deprecated_composite_definition_label
    def evaluate_failure_criteria(
        self,
        combined_criterion: CombinedFailureCriterion,
        composite_scope: Optional[CompositeScope] = None,
        measure: FailureMeasureEnum = FailureMeasureEnum.INVERSE_RESERVE_FACTOR,
        write_data_for_full_element_scope: bool = True,
        max_chunk_size: int = 50000,
    ) -> FieldsContainer:
        """Get a fields container with the evaluated failure criteria.

        The fields container contains the maximum per element if the measure
        is :attr:`.FailureMeasureEnum.INVERSE_RESERVE_FACTOR` and the minimum per element
        if the measure is the :attr:`.FailureMeasureEnum.MARGIN_OF_SAFETY` or
        :attr:`.FailureMeasureEnum.RESERVE_FACTOR` attribute.

        Parameters
        ----------
        combined_criterion :
            Combined failure criterion to evaluate.
        composite_scope :
            Composite scope on which to evaluate the failure criteria. If empty, the criteria
            is evaluated on the full model. If the time is not set, the last time or
            frequency in the result file is used.
        measure :
            Failure measure to evaluate.
        write_data_for_full_element_scope :
            Whether each element in the element scope is to get a
            (potentially zero) failure value, even elements that are not
            part of ``composite_scope.plies``. If no element scope is
            specified (``composite_scope.elements``), a (potentially zero)
            failure value is written for all elements.
        max_chunk_size:
            Maximum chunk size. If number of elements is larger than this number,
            the failure criteria are evaluated in chunks of the specified size.

            .. note::

                For some special element types such as beams,
                ``write_data_for_full_element_scope=True`` is not supported.

        """
        if composite_scope is None:
            composite_scope = CompositeScope()

        element_scope_in = [] if composite_scope.elements is None else composite_scope.elements
        ply_scope_in = [] if composite_scope.plies is None else composite_scope.plies
        ns_in = [] if composite_scope.named_selections is None else composite_scope.named_selections
        time_in = composite_scope.time

        if composite_scope.plies is None or len(composite_scope.plies):
            # This is a workaround because setting the
            # write_data_for_full_element_scope flag to True can lead to
            # problems with 2023 R1 if non-composite elements such as
            # beams exist in the solution. Because the flag
            # is irrelevant for cases without a ply scope, we set it to False here.
            write_data_for_full_element_scope = False

        has_layup_provider = len(self._composite_files.composite) > 0

        # configure primary scoping
        scope_config = dpf.DataTree()
        if time_in:
            scope_config.add({"requested_times": time_in})
        scope_config_reader_op = dpf.Operator("composite::scope_config_reader")
        scope_config_reader_op.inputs.scope_configuration(scope_config)

        if ply_scope_in:
            selected_plies_op = dpf.Operator("composite::string_container")
            for value in enumerate(ply_scope_in):
                selected_plies_op.connect(value[0], value[1])
            scope_config_reader_op.inputs.ply_ids(selected_plies_op.outputs.strings)

        # configure operator to chunk the scope
        chunking_data_tree = dpf.DataTree({"max_chunk_size": 50000})
        if ns_in:
            chunking_data_tree.add({"named_selections": ns_in})

        chunking_generator = dpf.Operator("composite::scope_generator")
        chunking_generator.inputs.stream_provider(self._get_rst_streams_provider())
        chunking_generator.inputs.data_tree(chunking_data_tree)
        chunking_generator.inputs.data_sources(self.data_sources.composite)
        if element_scope_in:
            element_scope = dpf.Scoping(location="elemental")
            element_scope.ids = element_scope_in
            chunking_generator.inputs.element_scoping(element_scope)

        min_merger = dpf.Operator("merge::fields_container")
        max_merger = dpf.Operator("merge::fields_container")

        merge_index = 0

        while True:
            chunking_generator.inputs.generator_counter(merge_index)
            finished = chunking_generator.outputs.is_finished()
            if finished:
                break

            evaluate_failure_criterion_per_scope_op = dpf.Operator(
                "composite::evaluate_failure_criterion_per_scope"
            )

            # Live evaluation is currently not supported by the Python module
            # because the docker container does not support it.
            evaluate_failure_criterion_per_scope_op.inputs.criterion_configuration(
                combined_criterion.to_json()
            )

            evaluate_failure_criterion_per_scope_op.inputs.scope_configuration(
                scope_config_reader_op.outputs
            )

            evaluate_failure_criterion_per_scope_op.inputs.element_scoping(
                chunking_generator.outputs
            )
            evaluate_failure_criterion_per_scope_op.inputs.materials_container(
                self.material_operators.material_provider.outputs
            )
            evaluate_failure_criterion_per_scope_op.inputs.stream_provider(
                self._get_rst_streams_provider()
            )
            evaluate_failure_criterion_per_scope_op.inputs.mesh(self.get_mesh())
            evaluate_failure_criterion_per_scope_op.inputs.has_layup_provider(has_layup_provider)
            evaluate_failure_criterion_per_scope_op.inputs.section_data_container(
                self._layup_provider.outputs.section_data_container
            )
            evaluate_failure_criterion_per_scope_op.inputs.material_fields(
                self._layup_provider.outputs.material_fields
            )
            evaluate_failure_criterion_per_scope_op.inputs.mesh_properties_container(
                self._layup_provider.outputs.mesh_properties_container
            )
            # Ensure that sandwich criteria are evaluated
            evaluate_failure_criterion_per_scope_op.inputs.request_sandwich_results(True)

            minmax_el_op = dpf.Operator("composite::minmax_per_element_operator")
            minmax_el_op.inputs.fields_container(
                evaluate_failure_criterion_per_scope_op.outputs.failure_container
            )

            minmax_el_op.inputs.mesh(self.get_mesh())
            minmax_el_op.inputs.material_support(
                self.material_operators.material_support_provider.outputs
            )

            if has_layup_provider and write_data_for_full_element_scope:
                add_default_data_op = dpf.Operator("composite::add_default_data")
                add_default_data_op.inputs.requested_element_scoping(chunking_generator.outputs)
                add_default_data_op.inputs.time_id(
                    evaluate_failure_criterion_per_scope_op.outputs.time_id
                )

                add_default_data_op.inputs.mesh(self.get_mesh())

                add_default_data_op.inputs.fields_container(minmax_el_op.outputs.field_min)
                add_default_data_op.run()

                add_default_data_op.inputs.fields_container(minmax_el_op.outputs.field_max)
                add_default_data_op.run()

            # It is important to evaluate the field here, otherwise the merge operator detects
            # the workflow as changed if upstream operator inputs change
            min_container = minmax_el_op.outputs.field_min()
            max_container = minmax_el_op.outputs.field_max()

            converter_op = dpf.Operator("composite::failure_measure_converter")
            converter_op.inputs.fields_container(min_container)
            converter_op.inputs.measure_type(measure.value)
            converter_op.run()

            converter_op.inputs.fields_container(max_container)
            converter_op.run()

            min_merger.connect(merge_index, min_container)
            max_merger.connect(merge_index, max_container)
            merge_index = merge_index + 1

        if merge_index == 0:
            raise RuntimeError(
                "No output is generated! Please check the scope (element and ply ids)."
            )

        if measure == FailureMeasureEnum.INVERSE_RESERVE_FACTOR:
            return max_merger.outputs.merged_fields_container()
        else:
            return min_merger.outputs.merged_fields_container()

    @_deprecated_composite_definition_label
    def get_sampling_point(
        self,
        combined_criterion: CombinedFailureCriterion,
        element_id: int,
        time: Optional[float] = None,
        composite_definition_label: Optional[str] = None,
    ) -> SamplingPointNew:
        """Get a sampling point for an element ID and failure criteria.

        Parameters
        ----------
        combined_criterion:
            Combined failure criterion to evaluate.
        element_id:
            Element ID or label of the sampling point.
        time:
            Time or frequency to evaluate the sampling point at. The default
            is ``None``, in which case the last time or frequency in the result
            file is used.
        composite_definition_label:
            Label of the composite definition, which is the
            dictionary key in the :attr:`.ContinuousFiberCompositesFiles.composite`
            attribute. This parameter is only required for assemblies.
            See the note about assemblies in the description for the :class:`CompositeModel` class.
        """
        time_in = time

        return SamplingPointNew(
            "Sampling Point",
            element_id,
            combined_criterion,
            self._material_operators,
            self.get_mesh(),
            self._layup_provider,
            self._get_rst_streams_provider(),
            self._data_sources.rst,
            self._unit_system,
            time_in,
        )

    @_deprecated_composite_definition_label
    def get_element_info(
        self, element_id: int, composite_definition_label: Optional[str] = None
    ) -> Optional[ElementInfo]:
        """Get element information for an element ID.

        This method returns ``None`` if the element type is not supported.

        Parameters
        ----------
        element_id:
            Element ID or label.
        composite_definition_label:
            Label of the composite definition, which is the
            dictionary key in the :attr:`.ContinuousFiberCompositesFiles.composite`
            attribute. This parameter is only required for assemblies.
            See the note about assemblies in the description for the :class:`CompositeModel` class.
        """
        return self._element_info_provider.get_element_info(element_id)

    @_deprecated_composite_definition_label
    def get_property_for_all_layers(
        self,
        layup_property: LayerProperty,
        element_id: int,
        composite_definition_label: Optional[str] = None,
    ) -> Optional[NDArray[np.double]]:
        """Get a layer property for an element ID.

        This method returns a numpy array with the values of the property for all the layers.
        Values are ordered from bottom to top.

        This method returns ``None`` if the element is not layered.

        Parameters
        ----------
        layup_property:
            Lay-up property.
        element_id:
            Element ID or label.
        composite_definition_label:
            Label of the composite definition, which is the
            dictionary key in the :attr:`.ContinuousFiberCompositesFiles.composite`
            attribute. This parameter is only required for assemblies.
            See the note about assemblies in the description for the :class:`CompositeModel` class.
        """
        if layup_property == LayerProperty.ANGLES:
            return self._layup_properties_provider.get_layer_angles(element_id)
        if layup_property == LayerProperty.THICKNESSES:
            return self._layup_properties_provider.get_layer_thicknesses(element_id)
        if layup_property == LayerProperty.SHEAR_ANGLES:
            return self._layup_properties_provider.get_layer_shear_angles(element_id)
        raise RuntimeError(f"Invalid property {layup_property}")

    @_deprecated_composite_definition_label
    def get_analysis_plies(
        self, element_id: int, composite_definition_label: Optional[str] = None
    ) -> Optional[Sequence[str]]:
        """Get analysis ply names.

        This method returns ``None`` if the element is not layered.

        Parameters
        ----------
        element_id:
            Element ID or label.
        composite_definition_label:
            Label of the composite definition, which is the
            dictionary key in the :attr:`.ContinuousFiberCompositesFiles.composite`
            attribute. This parameter is only required for assemblies.
            See the note about assemblies in the description for the :class:`CompositeModel` class.
            The dictionary only contains the analysis plies in the specified composite
            definition.
        """
        return self._layup_properties_provider.get_analysis_plies(element_id)

    @_deprecated_composite_definition_label
    def get_element_laminate_offset(
        self, element_id: int, composite_definition_label: Optional[str] = None
    ) -> Optional[np.double]:
        """Get the laminate offset of an element.

        This method returns ``None`` if the element is not layered.

        Parameters
        ----------
        element_id:
            Element ID or label.
        composite_definition_label:
            Label of the composite definition, which is the
            dictionary key in the :attr:`.ContinuousFiberCompositesFiles.composite`
            attribute. This parameter is only required for assemblies.
            See the note about assemblies in the description for the :class:`CompositeModel` class.
        """
        return self._layup_properties_provider.get_element_laminate_offset(element_id)

    @_deprecated_composite_definition_label
    def get_constant_property_dict(
        self,
        material_properties: Collection[MaterialProperty],
        composite_definition_label: Optional[str] = None,
    ) -> Dict[np.int64, Dict[MaterialProperty, float]]:
        """Get a dictionary with constant properties.

        This method returns a dictionary with ``dpf_material_id`` as the key and
        a dictionary with the requested properties as the value. Only constant properties
        are supported. Variable properties are evaluated at their
        default values.

        This method can be slow to evaluate and should not
        be called in a loop.

        Parameters
        ----------
        material_properties:
            List of the requested material properties.
        composite_definition_label:
            Label of the composite definition, which is the
            dictionary key in the :attr:`.ContinuousFiberCompositesFiles.composite`
            attribute. This parameter is only required for assemblies.
            See the note about assemblies in the description for the :class:`CompositeModel` class.
            The dictionary only contains the materials of the analysis plies defined
            in the specified composite definition.
        """
        return get_constant_property_dict(
            material_properties=material_properties,
            materials_provider=self.material_operators.material_provider,
            data_source_or_streams_provider=self._get_rst_streams_provider(),
            mesh=self.get_mesh(),
        )

    def get_result_times_or_frequencies(self) -> NDArray[np.double]:
        """Get the times or frequencies in the result file."""
        return cast(
            NDArray[np.double], self._core_model.metadata.time_freq_support.time_frequencies.data
        )

    @_deprecated_composite_definition_label
    def add_interlaminar_normal_stresses(
        self,
        stresses: FieldsContainer,
        strains: FieldsContainer,
        composite_definition_label: Optional[str] = None,
    ) -> None:
        """Add interlaminar normal stresses to the stresses fields container.

        For a usage example, see
        :ref:`sphx_glr_examples_gallery_examples_007_interlaminar_normal_stress_example.py`.

        Parameters
        ----------
        stresses:
            Stresses fields container to add interlaminar normal stresses to.
        strains:
            Strains fields container from which the interlaminar normal stresses
            are computed.

        composite_definition_label:
            Label of the composite definition, which is the
            dictionary key in the :attr:`.ContinuousFiberCompositesFiles.composite`
            attribute. This parameter is only required for assemblies.
            See the note about assemblies in the description for the :class:`CompositeModel` class.
            Interlaminar normal stresses are only added to the layered elements defined
            in the specified composite definition.
        """
        ins_operator = dpf.Operator("composite::interlaminar_normal_stress_operator")
        ins_operator.inputs.materials_container(self._material_operators.material_provider)
        ins_operator.inputs.mesh(self.get_mesh())
        ins_operator.inputs.mesh_properties_container(
            self._layup_provider.outputs.mesh_properties_container
        )
        ins_operator.inputs.section_data_container(
            self._layup_provider.outputs.section_data_container
        )
        ins_operator.inputs.strains_container(strains)
        ins_operator.inputs.stresses_container(stresses)

        # call run because ins operator has not output
        ins_operator.run()

    def get_all_layered_element_ids(self) -> Sequence[int]:
        """Get all layered element IDs."""
        return cast(
            List[int],
            self.get_mesh().property_field("element_layer_indices").scoping.ids,
        )

    def get_all_layered_element_ids_for_composite_definition_label(
        self, composite_definition_label: Optional[str] = None
    ) -> Sequence[int]:
        """Get all layered element IDs that belong to a composite definition label.

        Parameters
        ----------
        composite_definition_label:
            Deprecated. This parameter is no longer needed.
        """
        warn(
            "The get_all_layered_element_ids_for_composite_definition_label method is deprecated. "
            "Use the get_all_layered_element_ids method instead.",
            category=DeprecationWarning,
            stacklevel=2,
        )
        return self.get_all_layered_element_ids()

    def _get_rst_streams_provider(self) -> Operator:
        return self._core_model.metadata.streams_provider

    def _first_composite_definition_label_if_only_one(self) -> str:
        if len(self.composite_definition_labels) == 1:
            return self.composite_definition_labels[0]
        else:
            raise RuntimeError(
                f"Multiple composite definition keys exist: {self.composite_definition_labels}. "
                f"Specify a key explicitly."
            )



"""Module for the post-processing of composite structures."""

try:
    import importlib.metadata as importlib_metadata
except ModuleNotFoundError:
    import importlib_metadata  # type: ignore

__version__ = importlib_metadata.version(__name__.replace(".", "-"))

from . import (
    composite_model,
    composite_scope,
    constants,
    data_sources,
    failure_criteria,
    layup_info,
    result_definition,
    sampling_point,
    select_indices,
    server_helpers,
)

__all__ = (
    "composite_model",
    "composite_scope",
    "constants",
    "data_sources",
    "failure_criteria",
    "layup_info",
    "result_definition",
    "sampling_point",
    "server_helpers",
    "select_indices",
)



"""Composite Model Interface 2023R2."""
from typing import Collection, Dict, List, Optional, Sequence, cast
from warnings import warn

import ansys.dpf.core as dpf
from ansys.dpf.core import FieldsContainer, MeshedRegion, Operator, UnitSystem
from ansys.dpf.core.server_types import BaseServer
import numpy as np
from numpy.typing import NDArray

from .composite_scope import CompositeScope
from .data_sources import (
    CompositeDataSources,
    ContinuousFiberCompositesFiles,
    get_composites_data_sources,
)
from .failure_criteria import CombinedFailureCriterion
from .layup_info import (
    ElementInfo,
    LayerProperty,
    LayupPropertiesProvider,
    add_layup_info_to_mesh,
    get_element_info_provider,
)
from .layup_info.material_operators import MaterialOperators, get_material_operators
from .layup_info.material_properties import MaterialProperty, get_constant_property_dict
from .result_definition import FailureMeasureEnum, ResultDefinition, ResultDefinitionScope
from .sampling_point_2023r2 import SamplingPoint2023R2
from .sampling_point_types import SamplingPoint
from .server_helpers import upload_continuous_fiber_composite_files_to_server
from .unit_system import UnitSystemProvider, get_unit_system


class CompositeInfo:
    """Contains composite data providers for a composite definition."""

    def __init__(
        self,
        data_sources: CompositeDataSources,
        composite_definition_label: str,
        streams_provider: dpf.Operator,
        material_operators: MaterialOperators,
        unit_system: UnitSystemProvider,
    ):
        """Initialize ``CompositeInfo`` class and add enriched mesh with composite information."""
        mesh_provider = dpf.Operator("MeshProvider")
        mesh_provider.inputs.data_sources(data_sources.rst)
        self.mesh = mesh_provider.outputs.mesh()

        self.layup_provider = add_layup_info_to_mesh(
            mesh=self.mesh,
            data_sources=data_sources,
            material_operators=material_operators,
            unit_system=unit_system,
            composite_definition_label=composite_definition_label,
        )

        self.element_info_provider = get_element_info_provider(
            mesh=self.mesh,
            stream_provider_or_data_source=streams_provider,
        )
        self.layup_properties_provider = LayupPropertiesProvider(
            layup_provider=self.layup_provider, mesh=self.mesh
        )


class CompositeModelImpl2023R2:
    """Provides access to the basic composite postprocessing functionality.

    This class supports DPF Server version 6.1 (2-23 R2) and earlier.
    On initialization, the ``CompositeModel`` class automatically adds composite lay-up
    information to the meshed regions. It prepares the providers for different lay-up
    properties so that they can be efficiently evaluated. The composite_files provided
    are automatically uploaded to the server if needed.

    .. note::

        When creating a ``CompositeModel`` instance, several providers are created and
        lay-up information is added to the DPF meshed regions. Depending on the use
        case, it can be more efficient to create the providers separately.

        For assemblies with multiple composite definition files, separate meshes and
        lay-up operators are generated (wrapped by the ``CompositeInfo`` class). This
        is needed because the lay-up provider can currently only add the data of a single
        composite definitions file to a mesh. All functions that depend on composite
        definitions mut be called with the correct ``composite_definition_label``
        parameter. The layered elements that get information from a given
        composite definition label can be determined by calling
        ``self.get_all_layered_element_ids_for_composite_definition_label``.
        All the elements that are not part of a composite definition are either homogeneous
        solids or layered models defined outside of an ACP model. The
        ``self.composite_definition_labels`` command returns all available composite
        definition labels. For more information, see
        :ref:`sphx_glr_examples_gallery_examples_008_assembly_example.py`.

    Parameters
    ----------
    composite_files:
        Use the :func:`.get_composite_files_from_workbench_result_folder` function to obtain
        the :class:`.ContinuousFiberCompositesFiles` object.
    server:
        DPF Server on which the model is created
    default_unit_system:
        Unit system that is used if the result file
        does not specify the unit system. This happens
        for pure MAPDL projects.
    """

    def __init__(
        self,
        composite_files: ContinuousFiberCompositesFiles,
        server: BaseServer,
        default_unit_system: Optional[UnitSystem] = None,
    ):
        """Initialize data providers and add composite information to meshed region."""
        self._composite_files = upload_continuous_fiber_composite_files_to_server(
            composite_files, server
        )
        self._data_sources = get_composites_data_sources(self._composite_files)

        self._core_model = dpf.Model(self._data_sources.rst, server=server)
        self._server = server

        self._unit_system = get_unit_system(self._data_sources.rst, default_unit_system)

        self._material_operators = get_material_operators(
            rst_data_source=self._data_sources.material_support,
            unit_system=self._unit_system,
            engineering_data_source=self._data_sources.engineering_data,
        )

        self._composite_infos: Dict[str, CompositeInfo] = {}
        for composite_definition_label in self._data_sources.old_composite_sources:
            self._composite_infos[composite_definition_label] = CompositeInfo(
                data_sources=self._data_sources,
                composite_definition_label=composite_definition_label,
                streams_provider=self._core_model.metadata.streams_provider,
                material_operators=self._material_operators,
                unit_system=self._unit_system,
            )

        if len(self._composite_infos) > 1:
            warn(
                "The post-processing of composite models with multiple lay-up"
                " definitions (assemblies) is deprecated with DPF servers older than 7.0"
                " (2024 R1). Please use DPF server 7.0 or later.",
                DeprecationWarning,
                stacklevel=2,
            )

    @property
    def composite_definition_labels(self) -> Sequence[str]:
        """All composite definition labels in the model.

        This property is only relevant for assemblies.
        """
        return list(self._composite_infos.keys())

    @property
    def composite_files(self) -> ContinuousFiberCompositesFiles:
        """Get the composite file paths on the server."""
        return self._composite_files

    def get_mesh(self, composite_definition_label: Optional[str] = None) -> MeshedRegion:
        """Get the underlying DPF meshed region.

        The meshed region contains the lay-up information.

        Parameters
        ----------
        composite_definition_label :
            Label of the composite definition, which is the
            dictionary key in the :attr:`.ContinuousFiberCompositesFiles.composite`
            attribute. This parameter is only required for assemblies.
            See the note about assemblies in the description for the :class:`CompositeModel` class.
        """
        if composite_definition_label is None:
            composite_definition_label = self._first_composite_definition_label_if_only_one()
        return self._composite_infos[composite_definition_label].mesh

    @property
    def data_sources(self) -> CompositeDataSources:
        """Composite data sources."""
        return self._data_sources

    @property
    def core_model(self) -> dpf.Model:
        """Underlying DPF core model."""
        return self._core_model

    @property
    def material_operators(self) -> MaterialOperators:
        """Material operators."""
        return self._material_operators

    def get_layup_operator(self, composite_definition_label: Optional[str] = None) -> Operator:
        """Get the lay-up operator.

        Parameters
        ----------
        composite_definition_label :
            Label of the composite definition, which is the
            dictionary key in the :attr:`.ContinuousFiberCompositesFiles.composite`
            attribute. This parameter is only required for assemblies.
            See the note about assemblies in the description for the :class:`CompositeModel` class.

        """
        if composite_definition_label is None:
            composite_definition_label = self._first_composite_definition_label_if_only_one()
        return self._composite_infos[composite_definition_label].layup_provider

    def evaluate_failure_criteria(
        self,
        combined_criterion: CombinedFailureCriterion,
        composite_scope: Optional[CompositeScope] = None,
        measure: FailureMeasureEnum = FailureMeasureEnum.INVERSE_RESERVE_FACTOR,
        write_data_for_full_element_scope: bool = True,
    ) -> FieldsContainer:
        """Get a fields container with the evaluated failure criteria.

        The fields container contains the maximum per element if the measure
        is :attr:`.FailureMeasureEnum.INVERSE_RESERVE_FACTOR` and the minimum per element
        if the measure is :attr:`.FailureMeasureEnum.MARGIN_OF_SAFETY` or
        :attr:`.FailureMeasureEnum.RESERVE_FACTOR`.

        Parameters
        ----------
        combined_criterion :
            Combined failure criterion to evaluate.
        composite_scope :
            Composite scope on which to evaluate the failure criteria. If empty, the criteria
            is evaluated on the full model. If the time is not set, the last time or
            frequency in the result file is used.
        measure :
            Failure measure to evaluate.
        write_data_for_full_element_scope :
            Whether each element in the element scope is to get a
            (potentially zero) failure value, even elements that are not
            part of ``composite_scope.plies``. If no element scope is
            specified (``composite_scope.elements``), a (potentially zero)
            failure value is written for all elements.

            .. note::

                For some special element types such as beams,
                ``write_data_for_full_element_scope=True`` is not supported.

        """
        if composite_scope is None:
            composite_scope = CompositeScope()

        element_scope_in = [] if composite_scope.elements is None else composite_scope.elements
        ply_scope_in = [] if composite_scope.plies is None else composite_scope.plies
        ns_in = [] if composite_scope.named_selections is None else composite_scope.named_selections
        time_in = composite_scope.time

        if composite_scope.plies is None or len(composite_scope.plies):
            # This is a workaround because setting the
            # write_data_for_full_element_scope flag to True can lead to
            # problems with 2023 R1 if non-composite elements such as
            # beams exist in the solution. Because the flag
            # is irrelevant for cases without a ply scope, we set it to False here.
            write_data_for_full_element_scope = False

        scopes = []

        for composite_definition_label in self.composite_definition_labels:
            composite_files = self._composite_files.composite[composite_definition_label]
            scopes.append(
                ResultDefinitionScope(
                    composite_definition=composite_files.definition,
                    mapping_file=composite_files.mapping,
                    element_scope=element_scope_in,
                    ply_scope=ply_scope_in,
                    named_selection_scope=ns_in,
                    write_data_for_full_element_scope=write_data_for_full_element_scope,
                )
            )

        rd = ResultDefinition(
            name="combined failure criteria",
            rst_files=self._composite_files.rst,
            material_file=self._composite_files.engineering_data,
            combined_failure_criterion=combined_criterion,
            composite_scopes=scopes,
            time=time_in,
            measure=measure.value,
        )
        failure_operator = dpf.Operator("composite::composite_failure_operator")
        failure_operator.inputs.unit_system(self._unit_system)

        failure_operator.inputs.result_definition(rd.to_json())

        if measure == FailureMeasureEnum.INVERSE_RESERVE_FACTOR:
            return failure_operator.outputs.fields_containerMax()
        else:
            return failure_operator.outputs.fields_containerMin()

    def get_sampling_point(
        self,
        combined_criterion: CombinedFailureCriterion,
        element_id: int,
        time: Optional[float] = None,
        composite_definition_label: Optional[str] = None,
    ) -> SamplingPoint:
        """Get a sampling point for an element ID and failure criteria.

        Parameters
        ----------
        combined_criterion:
            Combined failure criterion to evaluate.
        element_id:
            Element ID or label of the sampling point.
        time:
            Time or frequency at which to evaluate the sampling point. If ``None``,
            the last time or frequency in the result file is used.
        composite_definition_label:
            Label of the composite definition, which is the
            dictionary key in the :attr:`.ContinuousFiberCompositesFiles.composite`
            attribute. This parameter is only required for assemblies.
            See the note about assemblies in the description for the :class:`CompositeModel` class.
        """
        time_in = time

        if composite_definition_label is None:
            # jvonrick: Jan 2023: We could also try to determine composite_definition_label
            # based on the element_id and
            # self.get_all_layered_element_ids_for_composite_definition_label.
            # But the element_id of the sampling point can be changed later
            # In this case it is complicated switch to the correct composite definition, so
            # it is probably better to make it explicit that the sampling point is tied to
            # a composite definition
            composite_definition_label = self._first_composite_definition_label_if_only_one()

        scope = ResultDefinitionScope(
            composite_definition=self._composite_files.composite[
                composite_definition_label
            ].definition,
            mapping_file=self._composite_files.composite[composite_definition_label].mapping,
            element_scope=[element_id],
            ply_scope=[],
            named_selection_scope=[],
        )
        rd = ResultDefinition(
            name="combined failure criteria",
            rst_files=self._composite_files.rst,
            material_file=self._composite_files.engineering_data,
            combined_failure_criterion=combined_criterion,
            time=time_in,
            composite_scopes=[scope],
        )

        return SamplingPoint2023R2("Sampling Point", rd, self._unit_system, server=self._server)

    def get_element_info(
        self, element_id: int, composite_definition_label: Optional[str] = None
    ) -> Optional[ElementInfo]:
        """Get element information for an element ID.

        This method returns ``None`` if the element type is not supported.

        Parameters
        ----------
        element_id:
            Element ID or label.
        composite_definition_label:
            Label of the composite definition, which is the
            dictionary key in the :attr:`.ContinuousFiberCompositesFiles.composite`
            attribute. This parameter is only required for assemblies.
            See the note about assemblies in the description for the :class:`CompositeModel` class.
        """
        if composite_definition_label is None:
            composite_definition_label = self._first_composite_definition_label_if_only_one()
        return self._composite_infos[
            composite_definition_label
        ].element_info_provider.get_element_info(element_id)

    def get_property_for_all_layers(
        self,
        layup_property: LayerProperty,
        element_id: int,
        composite_definition_label: Optional[str] = None,
    ) -> Optional[NDArray[np.double]]:
        """Get a layer property for an element ID.

        Returns a numpy array with the values of the property for all the layers.
        Values are ordered from bottom to top.

        This method returns ``None`` if the element is not layered.

        Parameters
        ----------
        layup_property:
            Lay-up property.
        element_id:
            Element ID or label.
        composite_definition_label:
            Label of the composite definition, which is the
            dictionary key in the :attr:`.ContinuousFiberCompositesFiles.composite`
            attribute. This parameter is only required for assemblies.
            See the note about assemblies in the description for the :class:`CompositeModel` class.
        """
        if composite_definition_label is None:
            composite_definition_label = self._first_composite_definition_label_if_only_one()
        layup_properties_provider = self._composite_infos[
            composite_definition_label
        ].layup_properties_provider
        if layup_property == LayerProperty.ANGLES:
            return layup_properties_provider.get_layer_angles(element_id)
        if layup_property == LayerProperty.THICKNESSES:
            return layup_properties_provider.get_layer_thicknesses(element_id)
        if layup_property == LayerProperty.SHEAR_ANGLES:
            return layup_properties_provider.get_layer_shear_angles(element_id)
        raise RuntimeError(f"Invalid property {layup_property}")

    def get_analysis_plies(
        self, element_id: int, composite_definition_label: Optional[str] = None
    ) -> Optional[Sequence[str]]:
        """Get analysis ply names.

        This method returns ``None`` if the element is not layered.

        Parameters
        ----------
        element_id:
            Element ID or label.
        composite_definition_label:
            Label of the composite definition, which is the
            dictionary key in the :attr:`.ContinuousFiberCompositesFiles.composite`
            attribute. This parameter is only required for assemblies.
            See the note about assemblies in the description for the :class:`CompositeModel` class.
            The dictionary only contains the analysis plies in the specified composite
            definition.
        """
        if composite_definition_label is None:
            composite_definition_label = self._first_composite_definition_label_if_only_one()
        layup_properties_provider = self._composite_infos[
            composite_definition_label
        ].layup_properties_provider
        return layup_properties_provider.get_analysis_plies(element_id)

    def get_element_laminate_offset(
        self, element_id: int, composite_definition_label: Optional[str] = None
    ) -> Optional[np.double]:
        """Get the laminate offset of an element.

        THis method returns ``None`` if the element is not layered.

        Parameters
        ----------
        element_id:
            Element ID or label.
        composite_definition_label:
            Label of the composite definition, which is the
            dictionary key in the :attr:`.ContinuousFiberCompositesFiles.composite`
            attribute. This parameter is only required for assemblies.
            See the note about assemblies in the description for the :class:`CompositeModel` class.
        """
        if composite_definition_label is None:
            composite_definition_label = self._first_composite_definition_label_if_only_one()
        layup_properties_provider = self._composite_infos[
            composite_definition_label
        ].layup_properties_provider
        return layup_properties_provider.get_element_laminate_offset(element_id)

    def get_constant_property_dict(
        self,
        material_properties: Collection[MaterialProperty],
        composite_definition_label: Optional[str] = None,
    ) -> Dict[np.int64, Dict[MaterialProperty, float]]:
        """Get a dictionary with constant properties.

        Returns a dictionary with ``dpf_material_id`` as the key and
        a dictionary with the requested properties as the value. Only constant properties
        are supported. Variable properties are evaluated at their
        default values.

        This method can be slow to evaluate and should not
        be called in a loop.

        Parameters
        ----------
        material_properties:
            List of the requested material properties.
        composite_definition_label:
            Label of the composite definition, which is the
            dictionary key in the :attr:`.ContinuousFiberCompositesFiles.composite`
            attribute. This parameter is only required for assemblies.
            See the note about assemblies in the description for the :class:`CompositeModel` class.
            The dictionary only contains the materials of the analysis plies defined
            in the specified composite definition.
        """
        if composite_definition_label is None:
            composite_definition_label = self._first_composite_definition_label_if_only_one()
        return get_constant_property_dict(
            material_properties=material_properties,
            materials_provider=self.material_operators.material_provider,
            data_source_or_streams_provider=self.core_model.metadata.streams_provider,
            mesh=self.get_mesh(composite_definition_label),
        )

    def get_result_times_or_frequencies(self) -> NDArray[np.double]:
        """Get the times or frequencies in the result file."""
        return cast(
            NDArray[np.double], self._core_model.metadata.time_freq_support.time_frequencies.data
        )

    def add_interlaminar_normal_stresses(
        self,
        stresses: FieldsContainer,
        strains: FieldsContainer,
        composite_definition_label: Optional[str] = None,
    ) -> None:
        """Add interlaminar normal stresses to the stresses fields container.

        For a usage example, see
        :ref:`sphx_glr_examples_gallery_examples_007_interlaminar_normal_stress_example.py`.

        Parameters
        ----------
        stresses:
            Stresses fields container to add interlaminar normal stresses to.
        strains:
        composite_definition_label:
            Label of the composite definition, which is the
            dictionary key in the :attr:`.ContinuousFiberCompositesFiles.composite`
            attribute. This parameter is only required for assemblies.
            See the note about assemblies in the description for the :class:`CompositeModel` class.
            Interlaminar normal stresses are only added to the layered elements defined
            in the specified composite definition.
        """
        if composite_definition_label is None:
            composite_definition_label = self._first_composite_definition_label_if_only_one()

        layup_provider = self._composite_infos[composite_definition_label].layup_provider
        ins_operator = dpf.Operator("composite::interlaminar_normal_stress_operator")
        ins_operator.inputs.materials_container(self._material_operators.material_provider)
        ins_operator.inputs.mesh(self.get_mesh(composite_definition_label))
        ins_operator.inputs.mesh_properties_container(
            layup_provider.outputs.mesh_properties_container
        )
        ins_operator.inputs.section_data_container(layup_provider.outputs.section_data_container)
        ins_operator.inputs.strains_container(strains)
        ins_operator.inputs.stresses_container(stresses)

        # call run because ins operator has not output
        ins_operator.run()

    def get_all_layered_element_ids(self) -> Sequence[int]:
        """Get all layered element IDs."""
        raise NotImplementedError(
            "get_all_layered_element_ids is not implemented"
            " for this version of DPF. Please upgrade to 7.0 (2024 R1)"
            " or later. Use get_all_layered_element_ids_for_composite_definition_label"
            " instead."
        )

    def get_all_layered_element_ids_for_composite_definition_label(
        self, composite_definition_label: Optional[str]
    ) -> Sequence[int]:
        """Get all layered element IDs that belong to a composite definition label.

        Parameters
        ----------
        composite_definition_label:
            Label of the composite definition, which is the
            dictionary key in the :attr:`.ContinuousFiberCompositesFiles.composite`
            attribute. This parameter is only required for assemblies.
            See the note about assemblies in the description for the :class:`CompositeModel` class.
        """
        if not composite_definition_label:
            raise RuntimeError(
                "The composite_definition_label must be set for this version of DPF server."
                " Or update the the latest version."
            )
        return cast(
            List[int],
            self.get_mesh(composite_definition_label)
            .property_field("element_layer_indices")
            .scoping.ids,
        )

    def _first_composite_definition_label_if_only_one(self) -> str:
        if len(self.composite_definition_labels) == 1:
            return self.composite_definition_labels[0]
        else:
            raise RuntimeError(
                f"Multiple composite definition keys exist: {self.composite_definition_labels}. "
                f"Specify a key explicitly."
            )



"""Helpers to get material properties."""
from enum import Enum
from typing import Collection, Dict, Set, Union, cast

from ansys.dpf.core import DataSources, MeshedRegion, Operator, types
import numpy as np

__all__ = (
    "MaterialProperty",
    "get_constant_property",
    "get_all_dpf_material_ids",
    "get_constant_property_dict",
)
from ..unit_system import UnitSystemProvider, get_unit_system
from ._layup_info import get_element_info_provider


class MaterialProperty(str, Enum):
    """Provides the material properties that are available."""

    Engineering_Constants_E1: str = "young_modulus_x_direction"
    Engineering_Constants_E2: str = "young_modulus_y_direction"
    Engineering_Constants_E3: str = "young_modulus_z_direction"
    Engineering_Constants_E: str = "young_modulus"
    Engineering_Constants_G12: str = "shear_modulus_xy"
    Engineering_Constants_G23: str = "shear_modulus_yz"
    Engineering_Constants_G13: str = "shear_modulus_xz"
    Engineering_Constants_nu12: str = "poisson_ratio_xy"
    Engineering_Constants_nu23: str = "poisson_ratio_yz"
    Engineering_Constants_nu13: str = "poisson_ratio_xz"
    Engineering_Constants_nu: str = "poisson_ratio"
    Strain_Limits_effective_strain: str = "von_mises"
    Tensile_Yield_Strength_effective_stress: str = "tensile_yield_strength"
    Strain_Limits_eXt: str = "strain_tensile_x_direction"
    Strain_Limits_eYt: str = "strain_tensile_y_direction"
    Strain_Limits_eZt: str = "strain_tensile_z_direction"
    Strain_Limits_eXc: str = "strain_compressive_x_direction"
    Strain_Limits_eYc: str = "strain_compressive_y_direction"
    Strain_Limits_eZc: str = "strain_compressive_z_direction"
    Strain_Limits_eSxy: str = "strain_shear_xy"
    Strain_Limits_eSyz: str = "strain_shear_yz"
    Strain_Limits_eSxz: str = "strain_shear_xz"
    Stress_Limits_Xt: str = "stress_tensile_x_direction"
    Stress_Limits_Yt: str = "stress_tensile_y_direction"
    Stress_Limits_Zt: str = "stress_tensile_z_direction"
    Stress_Limits_Xc: str = "stress_compressive_x_direction"
    Stress_Limits_Yc: str = "stress_compressive_y_direction"
    Stress_Limits_Zc: str = "stress_compressive_z_direction"
    Stress_Limits_Sxy: str = "stress_shear_xy"
    Stress_Limits_Syz: str = "stress_shear_yz"
    Stress_Limits_Sxz: str = "stress_shear_xz"
    Thermal_Conductivity_K: str = "thermal_conductivity_isotropic"
    Thermal_Conductivity_K1: str = "thermal_conductivity_x_direction"
    Thermal_Conductivity_K2: str = "thermal_conductivity_y_direction"
    Thermal_Conductivity_K3: str = "thermal_conductivity_z_direction"
    Specific_Heat_cp: str = "specific_heat"
    Thermal_Expansion_Coefficients_a: str = "coefficient_thermal_expansion"
    Thermal_Expansion_Coefficients_aX: str = "coefficient_thermal_expansion_x"
    Thermal_Expansion_Coefficients_aY: str = "coefficient_thermal_expansion_y"
    Thermal_Expansion_Coefficients_aZ: str = "coefficient_thermal_expansion_z"
    Fabric_Fiber_Angle_phi: str = "fabric_fiber_angle"
    Hill_Yield_Criterion_R11: str = "yield_stress_ratio_x_hill_criterion"
    Hill_Yield_Criterion_R22: str = "yield_stress_ratio_y_hill_criterion"
    Hill_Yield_Criterion_R33: str = "yield_stress_ratio_z_hill_criterion"
    Hill_Yield_Criterion_R12: str = "yield_stress_ratio_xy_hill_criterion"
    Hill_Yield_Criterion_R23: str = "yield_stress_ratio_yz_hill_criterion"
    Hill_Yield_Criterion_R13: str = "yield_stress_ratio_xz_hill_criterion"
    Tsai_Wu_Constant_xy: str = "coupling_coef_xy_tsai_wu"
    Tsai_Wu_Constant_yz: str = "coupling_coef_yz_tsai_wu"
    Tsai_Wu_Constant_xz: str = "coupling_coef_xz_tsai_wu"
    Puck_Constants_p_21_pos: str = "tensile_inclination_xz_puck_constants"
    Puck_Constants_p_21_neg: str = "compressive_inclination_xz_puck_constants"
    Puck_Constants_p_22_pos: str = "tensile_inclination_yz_puck_constants"
    Puck_Constants_p_22_neg: str = "compressive_inclination_yz_puck_constants"
    Puck_Constants_s: str = "degradation_parameter_s_puck_constants"
    Puck_Constants_m: str = "degradation_parameter_m_puck_constants"
    Puck_Constants_interface_weakening_factor: str = "interface_weakening_factor_puck_constants"
    Larc_Constants_fracture_angle_under_compression: str = (
        "fracture_angle_under_compression_larc_constants"
    )
    Larc_Constants_fracture_toughness_ratio: str = "fracture_toughness_ratio_larc_constants"
    Larc_Constants_fracture_toughness_mode_1: str = (
        "longitudinal_friction_coefficient_larc_constants"
    )
    Larc_Constants_fracture_toughness_mode_2: str = "transverse_friction_coefficient_larc_constants"


def get_constant_property(
    material_property: MaterialProperty,
    dpf_material_id: np.int64,
    materials_provider: Operator,
    unit_system: UnitSystemProvider,
) -> float:
    """Get a constant material property.

    Only constant properties are supported. Variable properties are evaluated at their
    default values.

    Parameters
    ----------
    material_property:
        Material property.
    dpf_material_id:
        DPF material ID.
    materials_provider:
        DPF Materials provider operator. This value is available from the
        :attr:`.CompositeModel.material_operators` attribute.
    unit_system:
    """
    material_property_field = Operator("eng_data::ans_mat_property_field_provider")
    material_property_field.inputs.materials_container(materials_provider)
    material_property_field.inputs.dpf_mat_id(int(dpf_material_id))
    material_property_field.inputs.property_name(material_property.value)

    material_property_field.inputs.unit_system_or_result_info(unit_system)
    properties = material_property_field.get_output(output_type=types.fields_container)
    assert len(properties) == 1, "Properties container has to have exactly one entry."
    assert len(properties[0].data) == 1, (
        "Properties field has to have exactly one entry."
        "Probably the property is not constant. "
        "Variable properties are currently not supported."
    )
    return cast(float, properties[0].data[0])


def get_all_dpf_material_ids(
    mesh: MeshedRegion, data_source_or_streams_provider: Union[DataSources, Operator]
) -> Collection[np.int64]:
    """Get all DPF material IDs.

    Parameters
    ----------
    mesh:
        DPF meshed region enriched with lay-up information.
    data_source_or_streams_provider:
        DPF data source or stream provider that contains the RST file.
    """
    element_info_provider = get_element_info_provider(mesh, data_source_or_streams_provider)
    all_material_ids: Set["np.int64"] = set()
    for element_id in mesh.elements.scoping.ids:
        element_info = element_info_provider.get_element_info(element_id)
        if element_info is not None:
            all_material_ids.update(element_info.dpf_material_ids)
    return all_material_ids


def get_constant_property_dict(
    material_properties: Collection[MaterialProperty],
    materials_provider: Operator,
    data_source_or_streams_provider: Union[DataSources, Operator],
    mesh: MeshedRegion,
) -> Dict[np.int64, Dict[MaterialProperty, float]]:
    """Get a dictionary with constant properties.

    Returns a dictionary with the DPF material ID as a key and
    a dictionary with the requested properties as the value. Only constant properties
    are supported. Variable properties are evaluated at their
    default values.
    Because this method can be slow to evaluate, it should not
    be called in a loop.

    Parameters
    ----------
    material_properties:
        Material properties to request.
    materials_provider:
        DPF Materials provider operator. This value is Available from the
        :attr:`.CompositeModel.material_operators` attribute.
    data_source_or_streams_provider:
        DPF data source or stream provider that contains the RST file.
    mesh:
        DPF meshed region enriched with lay-up information.
    """
    properties: Dict[np.int64, Dict[MaterialProperty, float]] = {}
    unit_system = get_unit_system(data_source_or_streams_provider)
    for dpf_material_id in get_all_dpf_material_ids(
        mesh=mesh, data_source_or_streams_provider=data_source_or_streams_provider
    ):
        properties[dpf_material_id] = {}
        for material_property in material_properties:
            constant_property = get_constant_property(
                material_property=material_property,
                dpf_material_id=dpf_material_id,
                materials_provider=materials_provider,
                unit_system=unit_system,
            )
            properties[dpf_material_id][material_property] = constant_property
    return properties



"""Cuntze failure criterion."""

import inspect
import math

from ._failure_criterion_base import FailureCriterionBase

_DOC_CFC = "Activates the failure evaluation regarding compression in fiber direction."
_DOC_CFT = "Activates the failure evaluation regarding tension in fiber direction."
_DOC_CMA = "Activates the failure evaluation of the matrix due to tension."
_DOC_CMB = "Activates the failure evaluation of the matrix due to compression."
_DOC_CMC = "Activates the failure evaluation of the matrix due to compression or shear."
_DOC_DIM = "Whether the 2D or 3D formulation of the criterion is used."
_DOC_WF_CFC = "Weighting factor of fiber failure due to compression (cfc)."
_DOC_WF_CFT = "Weighting factor of fiber failure due to tension (cft)."
_DOC_WF_CMA = "Weighting factor of matrix failure due to tension (cma)."
_DOC_WF_CMB = "Weighting factor of matrix failure due to compression (cmb)."
_DOC_WF_CMC = "Weighting factor of matrix failure due to compression or shear (cmc)."
_DOC_B21 = "In-plane shear friction coefficient. The default is ``0.2``."
_DOC_B32 = (
    "Out-of-plane shear friction coefficient. The default is ``1.3805``. The value depends "
    "on the fracture plane angle."
)
_DOC_MODE_INTERACTION_COEFF = "Mode interaction coefficient. The default is ``2.6``."
_DOC_FRACTURE_PLANE_ANGLE = (
    "Fracture plane angle in degrees. The default is ``53``. "
    "The value must be greater than 45. The value depends on the out-of-plane "
    "shear friction coefficient."
)


class CuntzeCriterion(FailureCriterionBase):
    """Provides the Cuntze criterion."""

    __doc__ = f"""Defines the Cuntze criterion for uni-directional orthotropic reinforced materials.

    Parameters
    ----------
    cfc:
        {_DOC_CFC}
    cft:
        {_DOC_CFT}
    cma:
        {_DOC_CMA}
    cmb:
        {_DOC_CMB}
    cmc:
        {_DOC_CMC}
    dim:
        {_DOC_DIM}
    wf_cfc:
        {_DOC_WF_CFC}
    wf_cft:
        {_DOC_WF_CFT}
    wf_cma:
        {_DOC_WF_CMA}
    wf_cmb:
        {_DOC_WF_CMB}
    wf_cmc:
        {_DOC_WF_CMC}
    b21:
        {_DOC_B21}
    b32:
        {_DOC_B32}
    fracture_plane_angle:
        {_DOC_FRACTURE_PLANE_ANGLE}
    mode_interaction_coeff:
        {_DOC_MODE_INTERACTION_COEFF}
    """

    def __init__(
        self,
        *,
        cfc: bool = True,
        cft: bool = True,
        cma: bool = True,
        cmb: bool = True,
        cmc: bool = True,
        dim: int = 2,
        wf_cfc: float = 1.0,
        wf_cft: float = 1.0,
        wf_cma: float = 1.0,
        wf_cmb: float = 1.0,
        wf_cmc: float = 1.0,
        b21: float = 0.2,
        b32: float = 1.3805,
        fracture_plane_angle: float = 53.0,
        mode_interaction_coeff: float = 2.6,
    ) -> None:
        """Construct a Cuntze failure criterion."""
        super().__init__(name="Cuntze", active=True)

        for attr in ATTRS_CUNTZE:
            setattr(self, attr, eval(attr))

    def _get_cfc(self) -> bool:
        return self._cfc

    def _set_cfc(self, value: bool) -> None:
        self._cfc = value

    def _get_cft(self) -> bool:
        return self._cft

    def _set_cft(self, value: bool) -> None:
        self._cft = value

    def _get_cma(self) -> bool:
        return self._cma

    def _set_cma(self, value: bool) -> None:
        self._cma = value

    def _get_cmb(self) -> bool:
        return self._cmb

    def _set_cmb(self, value: bool) -> None:
        self._cmb = value

    def _get_cmc(self) -> bool:
        return self._cmc

    def _set_cmc(self, value: bool) -> None:
        self._cmc = value

    def _get_dim(self) -> int:
        return self._dim

    def _set_dim(self, value: int) -> None:
        if value in [2, 3]:
            self._dim = value
        else:
            raise AttributeError(
                f"Dimension of {self.name} cannot be set to {value}. The values allowed are 2 or 3."
            )

    def _get_wf_cfc(self) -> float:
        return self._wf_cfc

    def _set_wf_cfc(self, value: float) -> None:
        self._wf_cfc = value

    def _get_wf_cft(self) -> float:
        return self._wf_cft

    def _set_wf_cft(self, value: float) -> None:
        self._wf_cft = value

    def _get_wf_cma(self) -> float:
        return self._wf_cma

    def _set_wf_cma(self, value: float) -> None:
        self._wf_cma = value

    def _get_wf_cmb(self) -> float:
        return self._wf_cmb

    def _set_wf_cmb(self, value: float) -> None:
        self._wf_cmb = value

    def _get_wf_cmc(self) -> float:
        return self._wf_cmc

    def _set_wf_cmc(self, value: float) -> None:
        self._wf_cmc = value

    def _get_b21(self) -> float:
        return self._b21

    def _set_b21(self, value: float) -> None:
        self._b21 = value

    def _get_b32(self) -> float:
        return self._b32

    def _set_b32(self, value: float) -> None:
        if value >= 1.0:
            self._fracture_plane_angle = math.acos(1.0 / value - 1.0) / 2.0 * 180.0 / math.pi
            self._b32 = value
        else:
            raise ValueError("Out-of-plane friction coefficient b32 must be 1 or greater.")

    def _get_fracture_plane_angle(self) -> float:
        return self._fracture_plane_angle

    def _set_fracture_plane_angle(self, value: float) -> None:
        if value > 45.0:
            self._b32 = 1.0 / (1.0 + math.cos(2.0 * math.pi * value / 180.0))
            self._fracture_plane_angle = value
        else:
            raise ValueError("Fracture plane angle must be above 45 degrees.")

    def _get_mode_interaction_coeff(self) -> float:
        return self._mode_interaction_coeff

    def _set_mode_interaction_coeff(self, value: float) -> None:
        self._mode_interaction_coeff = value

    cfc = property(_get_cfc, _set_cfc, doc=_DOC_CFC)
    cft = property(_get_cft, _set_cft, doc=_DOC_CFT)
    cma = property(_get_cma, _set_cma, doc=_DOC_CMA)
    cmb = property(_get_cmb, _set_cmb, doc=_DOC_CMB)
    cmc = property(_get_cmc, _set_cmc, doc=_DOC_CMC)
    dim = property(_get_dim, _set_dim, doc=_DOC_DIM)
    wf_cfc = property(_get_wf_cfc, _set_wf_cfc, doc=_DOC_WF_CFC)
    wf_cft = property(_get_wf_cft, _set_wf_cft, doc=_DOC_WF_CFT)
    wf_cma = property(_get_wf_cma, _set_wf_cma, doc=_DOC_WF_CMA)
    wf_cmb = property(_get_wf_cmb, _set_wf_cmb, doc=_DOC_WF_CMB)
    wf_cmc = property(_get_wf_cmc, _set_wf_cmc, doc=_DOC_WF_CMC)
    b21 = property(_get_b21, _set_b21, doc=_DOC_B21)
    b32 = property(_get_b32, _set_b32, doc=_DOC_B32)
    mode_interaction_coeff = property(
        _get_mode_interaction_coeff,
        _set_mode_interaction_coeff,
        doc=_DOC_MODE_INTERACTION_COEFF,
    )
    fracture_plane_angle = property(
        _get_fracture_plane_angle,
        _set_fracture_plane_angle,
        doc=_DOC_FRACTURE_PLANE_ANGLE,
    )


ATTRS_CUNTZE = inspect.signature(CuntzeCriterion).parameters.keys()



"""Helpers to connect to or start a DPF server with the DPF Composites plugin."""
import os
from typing import Any, Callable, Dict, Optional, Union

from ansys.dpf.core import connect_to_server
from ansys.dpf.core import server as _dpf_server
from ansys.dpf.core import server_context, start_local_server

from ansys.dpf.composites.server_helpers._load_plugin import load_composites_plugin


def _try_until_timeout(fun: Callable[[], Any], error_message: str, timeout: int = 10) -> Any:
    """Try to run a function until a timeout is reached.

    Before the timeout is reached, all exceptions are ignored and a retry happens.
    """
    import time

    tstart = time.time()
    while (time.time() - tstart) < timeout:
        time.sleep(0.001)
        try:
            return fun()
        except Exception:  # pylint: disable=broad-except
            pass
    raise TimeoutError(f"Timeout is reached: {error_message}")


def _wait_until_server_is_up(server: _dpf_server) -> Any:
    # Small hack to check if the server is up.
    # The DPF server should check this in the ``connect_to_server`` function, but
    # that's currently not the case.
    # https://github.com/ansys/pydpf-core/issues/414
    # We use the fact that server.version throws an error if the server
    # is not yet connected.
    _try_until_timeout(
        lambda: server.version, "Failed to connect to the DPF server before timing out."
    )


def connect_to_or_start_server(
    port: Optional[int] = None, ip: Optional[str] = None, ansys_path: Optional[str] = None
) -> Any:
    r"""Connect to or start a DPF server with the DPF Composites plugin loaded.

    .. note::

        If a port or IP address is set, this method tries to connect to the server specified
        and the ``ansys_path`` parameter is ignored. If no parameters are set, a local server
        from the latest available Ansys installation is started.

    Parameters
    ----------
    port :
        Port that the DPF server is listening on.
    ip :
        IP address for the DPF server.
    ansys_path :
        Root path for the Ansys installation. For example, ``C:\\Program Files\\ANSYS Inc\\v232``.
        This parameter is ignored if either the port or IP address is set.

    Returns
    -------
    :
        DPF server.
    """
    port_in_env = os.environ.get("PYDPF_COMPOSITES_DOCKER_CONTAINER_PORT")
    if port_in_env is not None:
        port = int(port_in_env)

    connect_kwargs: Dict[str, Union[int, str]] = {}
    if port is not None:
        connect_kwargs["port"] = port
    if ip is not None:
        connect_kwargs["ip"] = ip

    if len(list(connect_kwargs.keys())) > 0:
        server = connect_to_server(
            **connect_kwargs,
            context=_dpf_server.server_context.ServerContext(
                server_context.LicensingContextType.premium
            ),
        )
    else:
        server = start_local_server(
            ansys_path=ansys_path,
            context=_dpf_server.server_context.ServerContext(
                server_context.LicensingContextType.premium
            ),
        )

    required_version = "6.0"
    server.check_version(
        required_version,
        f"The DPF Composites plugin requires DPF Server version {required_version} "
        f"(Ansys 2023 R2) or later. Your version is currently {server.version}.",
    )

    _wait_until_server_is_up(server)
    # Note: server.ansys_path contains the computed Ansys path from
    # dpf.server.start_local_server. It is None if
    # a connection is made to an existing server.
    load_composites_plugin(server, ansys_path=server.ansys_path)
    return server



"""Provides methods for reading the composite lay-up information."""

from . import material_operators, material_properties
from ._add_layup_info_to_mesh import add_layup_info_to_mesh
from ._enums import LayerProperty, LayupProperty
from ._layup_info import (
    AnalysisPlyInfoProvider,
    ElementInfo,
    ElementInfoProvider,
    LayupPropertiesProvider,
    get_all_analysis_ply_names,
    get_analysis_ply_index_to_name_map,
    get_dpf_material_id_by_analyis_ply_map,
    get_element_info_provider,
)

__all__ = (
    "add_layup_info_to_mesh",
    "AnalysisPlyInfoProvider",
    "ElementInfo",
    "ElementInfoProvider",
    "LayerProperty",
    "LayupProperty",
    "LayupPropertiesProvider",
    "get_all_analysis_ply_names",
    "get_analysis_ply_index_to_name_map",
    "get_dpf_material_id_by_analyis_ply_map",
    "get_element_info_provider",
    "material_properties",
    "material_operators",
)



"""LaRC failure criterion."""
import inspect

from ._failure_criterion_base import FailureCriterionBase

_DOC_LFC = "Whether to activate the failure evaluation regarding compression in fiber direction."

_DOC_LFT = "Whether to activate the failure evaluation regarding tension in fiber direction."

_DOC_LMC = "Whether to activate the failure evaluation of the matrix due to compression."

_DOC_LMT = "Whether to activate the failure evaluation of the matrix due to tension."

_DOC_DIM = (
    "Whether the 2D or 3D formulation of the criterion is used. 2D is equivalent to "
    "LaRC03, and 3D is equivalent to LaRC04."
)
_DOC_WF_LFC = "Weighting factor of fiber failure due to compression (lfc)."
_DOC_WF_LFT = "Weighting factor of fiber failure due to tension (lft)."
_DOC_WF_LMC = "Weighting factor of matrix failure due to compression (lmc)."
_DOC_WF_LMT = "Weighting factor of matrix failure due to tension (lmt)."


class LaRCCriterion(FailureCriterionBase):
    """LaRC Criterion."""

    __doc__ = f"""Defines the LaRC failure criterion for UD reinforced materials.

    Parameters
    ----------
    lft:
        {_DOC_LFT}
    lfc:
        {_DOC_LFC}
    lmt:
        {_DOC_LMT}
    lmc:
        {_DOC_LMC}
    dim:
        {_DOC_DIM}
    wf_lft:
        {_DOC_WF_LFT}
    wf_lfc:
        {_DOC_WF_LFC}
    wf_lmt:
        {_DOC_WF_LMT}
    wf_lmc:
        {_DOC_WF_LMC}
    """

    def __init__(
        self,
        *,
        lft: bool = True,
        lfc: bool = True,
        lmt: bool = True,
        lmc: bool = True,
        dim: int = 2,
        wf_lft: float = 1.0,
        wf_lfc: float = 1.0,
        wf_lmt: float = 1.0,
        wf_lmc: float = 1.0,
    ):
        """Create a LaRC failure criterion for uni-directional reinforced materials."""
        super().__init__(name="LaRC", active=True)

        for attr in ATTRS_LARC:
            setattr(self, attr, eval(attr))

    def _get_lfc(self) -> bool:
        return self._lfc

    def _set_lfc(self, value: bool) -> None:
        self._lfc = value

    def _get_lft(self) -> bool:
        return self._lft

    def _set_lft(self, value: bool) -> None:
        self._lft = value

    def _get_lmc(self) -> bool:
        return self._lmc

    def _set_lmc(self, value: bool) -> None:
        self._lmc = value

    def _get_lmt(self) -> bool:
        return self._lmt

    def _set_lmt(self, value: bool) -> None:
        self._lmt = value

    def _get_dim(self) -> int:
        return self._dim

    def _set_dim(self, value: int) -> None:
        if value in [2, 3]:
            self._dim = value
        else:
            raise AttributeError(
                f"Dimension of {self.name} cannot be set to {value}. Allowed are 2 or 3 for "
                f" LaRC03 and LaRC04 respectively."
            )

    def _get_wf_lfc(self) -> float:
        return self._wf_lfc

    def _set_wf_lfc(self, value: float) -> None:
        self._wf_lfc = value

    def _get_wf_lft(self) -> float:
        return self._wf_lft

    def _set_wf_lft(self, value: float) -> None:
        self._wf_lft = value

    def _get_wf_lmc(self) -> float:
        return self._wf_lmc

    def _set_wf_lmc(self, value: float) -> None:
        self._wf_lmc = value

    def _get_wf_lmt(self) -> float:
        return self._wf_lmt

    def _set_wf_lmt(self, value: float) -> None:
        self._wf_lmt = value

    lfc = property(
        _get_lfc,
        _set_lfc,
        doc=_DOC_LFC,
    )
    lft = property(
        _get_lft,
        _set_lft,
        doc=_DOC_LFT,
    )
    lmc = property(_get_lmc, _set_lmc, doc=_DOC_LMC)
    lmt = property(_get_lmt, _set_lmt, doc=_DOC_LMT)

    dim = property(
        _get_dim,
        _set_dim,
        doc=_DOC_DIM,
    )

    wf_lfc = property(_get_wf_lfc, _set_wf_lfc, doc=_DOC_WF_LFC)

    wf_lft = property(_get_wf_lft, _set_wf_lft, doc=_DOC_WF_LFT)

    wf_lmc = property(_get_wf_lmc, _set_wf_lmc, doc=_DOC_WF_LMC)
    wf_lmt = property(_get_wf_lmt, _set_wf_lmt, doc=_DOC_WF_LMT)


ATTRS_LARC = inspect.signature(LaRCCriterion).parameters.keys()



"""Indexer helper classes."""
from dataclasses import dataclass
from typing import Optional, Protocol, cast

from ansys.dpf.core import Field, PropertyField, Scoping
import numpy as np
from numpy.typing import NDArray


@dataclass(frozen=True)
class IndexToId:
    """Mapping maps id to index."""

    mapping: NDArray[np.int64]
    max_id: int


def setup_index_by_id(scoping: Scoping) -> IndexToId:
    """Create array that can be indexed by id to get the index.

    For ids which are not present in the scoping the array has a value of -1

    Parameters
    ----------
    scoping:
        DPF scoping
    """
    indices: NDArray[np.int64] = np.full(max(scoping.ids) + 1, -1, dtype=np.int64)
    indices[scoping.ids] = np.arange(len(scoping.ids))
    return IndexToId(mapping=indices, max_id=len(indices) - 1)


class PropertyFieldIndexerSingleValue(Protocol):
    """Protocol for single value property field indexer."""

    def by_id(self, entity_id: int) -> Optional[np.int64]:
        """Get index by id."""


class PropertyFieldIndexerArrayValue(Protocol):
    """Protocol for array valued property field indexer."""

    def by_id(self, entity_id: int) -> Optional[NDArray[np.int64]]:
        """Get index by id."""


# General comment for all Indexer:
# The .data call accesses the actual data. This sends the data over grpc which takes some time
# It looks like it returns a DpfArray for non-local fields and an numpy array for local fields.
# Without converting the DpfArray to a numpy array,
# performance during the lookup is about 50% slower.
# It is not clear why. To be checked with dpf team. If this is a local field there is no
# performance difference because the local field implementation already returns a numpy
# array


class PropertyFieldIndexerNoDataPointer:
    """Indexer for a property field with no data pointer."""

    def __init__(self, field: PropertyField):
        """Create indexer and get data."""
        index_by_id = setup_index_by_id(field.scoping)
        self._indices = index_by_id.mapping
        self._max_id = index_by_id.max_id
        self._data: NDArray[np.int64] = np.array(field.data, dtype=np.int64)

    def by_id(self, entity_id: int) -> Optional[np.int64]:
        """Get index by id.

        Parameters
        ----------
        entity_id
        """
        if entity_id > self._max_id:
            return None

        idx = self._indices[entity_id]
        if idx < 0:
            return None
        return cast(np.int64, self._data[idx])


class FieldIndexerNoDataPointer:
    """Indexer for a dpf field with no data pointer."""

    def __init__(self, field: Field):
        """Create indexer and get data."""
        index_by_id = setup_index_by_id(field.scoping)
        self._indices = index_by_id.mapping
        self._max_id = index_by_id.max_id
        self._data: NDArray[np.double] = np.array(field.data, dtype=np.double)

    def by_id(self, entity_id: int) -> Optional[np.double]:
        """Get index by id.

        Parameters
        ----------
        entity_id
        """
        if entity_id > self._max_id:
            return None
        idx = self._indices[entity_id]
        if idx < 0:
            return None
        return cast(np.double, self._data[idx])


class PropertyFieldIndexerNoDataPointerNoBoundsCheck:
    """Indexer for a property field with no data pointer and no bounds checks."""

    def __init__(self, field: PropertyField):
        """Create indexer and get data."""
        index_by_id = setup_index_by_id(field.scoping)
        self._indices = index_by_id.mapping
        self._data: NDArray[np.int64] = np.array(field.data, dtype=np.int64)

    def by_id(self, entity_id: int) -> Optional[np.int64]:
        """Get index by id.

        Parameters
        ----------
        entity_id
        """
        return cast(np.int64, self._data[self._indices[entity_id]])


class PropertyFieldIndexerWithDataPointer:
    """Indexer for a property field with data pointer."""

    def __init__(self, field: PropertyField):
        """Create indexer and get data."""
        index_by_id = setup_index_by_id(field.scoping)
        self._indices = index_by_id.mapping
        self._max_id = index_by_id.max_id

        self._data: NDArray[np.int64] = np.array(field.data, dtype=np.int64)
        self._n_components = field.component_count

        self._data_pointer: NDArray[np.int64] = np.append(
            field._data_pointer, len(self._data) * self._n_components
        )

    def by_id(self, entity_id: int) -> Optional[NDArray[np.int64]]:
        """Get index by id.

        Parameters
        ----------
        entity_id
        """
        if entity_id > self._max_id:
            return None

        idx = self._indices[entity_id]
        if idx < 0:
            return None
        return cast(
            NDArray[np.int64],
            self._data[
                self._data_pointer[idx]
                // self._n_components : self._data_pointer[idx + 1]
                // self._n_components
            ],
        )


class FieldIndexerWithDataPointer:
    """Indexer for a dpf field with data pointer."""

    def __init__(self, field: Field):
        """Create indexer and get data."""
        index_by_id = setup_index_by_id(field.scoping)
        self._indices = index_by_id.mapping
        self._max_id = index_by_id.max_id

        self._data: NDArray[np.double] = np.array(field.data, dtype=np.double)
        self._n_components = field.component_count

        self._data_pointer: NDArray[np.int64] = np.append(
            field._data_pointer, len(self._data) * self._n_components
        )

    def by_id(self, entity_id: int) -> Optional[NDArray[np.double]]:
        """Get index by id.

        Parameters
        ----------
        entity_id
        """
        if entity_id > self._max_id:
            return None

        idx = self._indices[entity_id]
        if idx < 0:
            return None
        return cast(
            NDArray[np.double],
            self._data[
                self._data_pointer[idx]
                // self._n_components : self._data_pointer[idx + 1]
                // self._n_components
            ],
        )


class PropertyFieldIndexerWithDataPointerNoBoundsCheck:
    """Indexer for a property field with data pointer and no bounds checks."""

    def __init__(self, field: PropertyField):
        """Create indexer and get data."""
        index_by_id = setup_index_by_id(field.scoping)
        self._indices = index_by_id.mapping

        self._data: NDArray[np.int64] = np.array(field.data, dtype=np.int64)

        self._n_components = field.component_count

        self._data_pointer: NDArray[np.int64] = np.append(
            field._data_pointer, len(self._data) * self._n_components
        )

    def by_id(self, entity_id: int) -> Optional[NDArray[np.int64]]:
        """Get index by id.

        Parameters
        ----------
        entity_id
        """
        idx = self._indices[entity_id]
        if idx < 0:
            return None
        return cast(
            NDArray[np.int64],
            self._data[
                self._data_pointer[idx]
                // self._n_components : self._data_pointer[idx + 1]
                // self._n_components
            ],
        )



"""Helper functions to add lay-up information to a DPF meshed region."""
from typing import Optional
from warnings import warn

from ansys.dpf.core import DataSources, MeshedRegion, Operator

from ..data_sources import CompositeDataSources
from ..unit_system import UnitSystemProvider
from .material_operators import MaterialOperators


def _get_composite_data_sources_for_layup_provider(
    data_sources: CompositeDataSources, composite_definition_label: Optional[str] = None
) -> DataSources:
    """
    Prepare DataSources of composite for the DPF server.

    Ensure that DataSources object is compatible with the version of the layup provider.
    """
    # import is here because of circular dependencies
    from ..server_helpers import version_older_than

    # pylint: disable=protected-access
    if version_older_than(data_sources.composite._server, "7.0"):
        if len(data_sources.old_composite_sources) == 1:
            return data_sources.composite
        else:
            if not composite_definition_label:
                raise RuntimeError(
                    "Calling _get_composite_data_sources_for_layup_provider"
                    " without composite_definition_label is not allowed."
                )

            if composite_definition_label in data_sources.old_composite_sources.keys():
                return data_sources.old_composite_sources[composite_definition_label]
            else:
                key_strs = ", ".join(data_sources.old_composite_sources.keys())
                raise RuntimeError(
                    f"Invalid key {composite_definition_label}."
                    " Available keys in composite data sources are"
                    f" {key_strs}."
                )

    return data_sources.composite


def add_layup_info_to_mesh(
    data_sources: CompositeDataSources,
    material_operators: MaterialOperators,
    mesh: MeshedRegion,
    unit_system: Optional[UnitSystemProvider] = None,
    composite_definition_label: Optional[str] = None,
) -> Operator:
    """Add lay-up information to the mesh.

    Creates a lay-up provider operator that is run and returned.

    Parameters
    ----------
    data_sources:
        DPF data sources available from the :attr:`.CompositeModel.data_sources` attribute.
    mesh:
        DPF meshed region available from the :meth:`.CompositeModel.get_mesh` method.
    material_operators:
       MaterialOperators object available from the :attr:`.CompositeModel.material_operators`
       attribute.
    unit_system:
        Unit system specification
    composite_definition_label:
        Label of the composite definition, which is the
        dictionary key in the :attr:`.ContinuousFiberCompositesFiles.composite`
        attribute. This parameter is only required for assemblies.
        See the note about assemblies in the description for the :class:`.CompositeModel` class.
    Returns
    -------
    :
        Lay-up provider operator.
    """
    composite_data_sources = _get_composite_data_sources_for_layup_provider(
        data_sources, composite_definition_label
    )

    # Set up the lay-up provider.
    # Reads the composite definition file and enriches the mesh
    # with the composite lay-up information.
    layup_provider = Operator("composite::layup_provider_operator")
    layup_provider.inputs.mesh(mesh)
    layup_provider.inputs.data_sources(composite_data_sources)
    layup_provider.inputs.abstract_field_support(
        material_operators.material_support_provider.outputs.abstract_field_support
    )

    if unit_system is None:
        warn(
            "Calling add_layup_info_to_mesh"
            "without a unit system is deprecated. Use get_unit_system"
            "to obtain the unit system.",
            DeprecationWarning,
            stacklevel=2,
        )
        unit_system = material_operators.result_info_provider

    layup_provider.inputs.unit_system(unit_system)
    layup_provider.run()

    return layup_provider



"""Puck Failure Criterion."""
import inspect

from ._failure_criterion_base import FailureCriterionBase

_DOC_PF = "Whether to activate the failure evaluation regarding fiber failure."
_DOC_PMA = "Whether to activate the failure evaluation regarding matrix failure due to tension."
_DOC_PMB = "Whether to activate the failure evaluation regarding matrix failure due to compression."
_DOC_PMC = (
    "Whether to activate the failure evaluation regarding matrix failure due to shear failure."
)
_DOC_PD = "Whether to activate the failure evaluation regarding delamination if dim is equal to 3."
_DOC_DIM = (
    "Whether the 2D or 3D formulation of the criterion is used. The latter one also "
    "supports the failure mode delamination. Use ``1`` for a simplified Puck criterion."
)
_DOC_WF_PF = "Weighting factor of the fiber failure (pf) mode."
_DOC_WF_PMA = "Weighting factor of the matrix failure (pma) mode."
_DOC_WF_PMB = "Weighting factor of the matrix failure (pmb) mode."
_DOC_WF_PMC = "Weighting factor of the matrix failure (pmc) mode."
_DOC_WF_PD = "Weighting factor of the delamination failure (pd) mode."
_DOC_CFPS = "Whether to consider inter-fiber strength reduction due to fiber parallel stresses."
_DOC_S = (
    "Specifies the effect of fiber parallel stresses regarding inter-fiber failure. "
    "s = [0, 1) and the default is ``0.5``."
)
_DOC_M = (
    "Specifies the effect of fiber parallel stresses regarding inter-fiber failure. "
    "M = [0, 1). The default is ``0.5``."
)
_DOC_INTERFACE_WEAKENING_FACTOR = (
    "Multiplicator for the interlaminar strength if "
    "failure mode ``pd`` is active. The default is ``0.8``."
)
_DOC_FORCE_GLOBAL_CONSTANTS = (
    "Whether to use global constants instead of the material-wise properties."
)
_DOC_P21_NEG = (
    "Global inclination factor in the"
    r" :math:`\sigma_1 - \tau_{12}` plane at :math:`\sigma_2 = 0` for compression."
)
_DOC_P21_POS = (
    "Global inclination factor in the"
    r" :math:`\sigma_1 - \tau_{12}` plane at :math:`\sigma_2 = 0` for tension."
)
_DOC_P22_NEG = (
    r"Global inclination factor of the fracture  plane :math:`\perp \perp` for compression."
)
_DOC_P22_POS = r"Global inclination factor of the fracture plane :math:`\perp \perp` for tension."


class PuckCriterion(FailureCriterionBase):
    """Puck Criterion."""

    __doc__ = f"""Defines the Puck failure criterion for UD reinforced materials.

    Parameters
    ----------
    pf:
        {_DOC_PF}
    pma:
        {_DOC_PMA}
    pmb:
        {_DOC_PMB}
    pmc:
        {_DOC_PMC}
    pd:
        {_DOC_PD}
    dim:
        {_DOC_DIM}
    wf_pf:
        {_DOC_WF_PF}
    wf_pma:
        {_DOC_WF_PMA}
    wf_pmb:
        {_DOC_WF_PMB}
    wf_pmc:
        {_DOC_WF_PMC}
    wf_pd:
        {_DOC_WF_PD}
    cfps:
        {_DOC_CFPS}
    s:
        {_DOC_S}
    m:
        {_DOC_M}
    interface_weakening_factor:
        {_DOC_INTERFACE_WEAKENING_FACTOR}
    force_global_constants:
        {_DOC_FORCE_GLOBAL_CONSTANTS}
    p21_neg:
        {_DOC_P21_NEG}
    p21_pos:
        {_DOC_P21_POS},
    p22_neg:
        {_DOC_P22_NEG}
    p22_pos:
        {_DOC_P22_POS}
    """

    def __init__(
        self,
        *,
        pf: bool = True,
        pma: bool = True,
        pmb: bool = True,
        pmc: bool = True,
        pd: bool = False,
        dim: int = 2,
        wf_pf: float = 1.0,
        wf_pma: float = 1.0,
        wf_pmb: float = 1.0,
        wf_pmc: float = 1.0,
        wf_pd: float = 1.0,
        cfps: bool = True,
        s: float = 0.5,
        M: float = 0.5,
        interface_weakening_factor: float = 0.8,
        force_global_constants: bool = False,
        p21_neg: float = 0.275,
        p21_pos: float = 0.325,
        p22_neg: float = 0.225,
        p22_pos: float = 0.225,
    ):
        """Create a Puck failure criterion.

        Can be used in combination with uni-directional orthotropic reinforced materials.
        """
        super().__init__(name="Puck", active=True)

        for attr in ATTRS_PUCK:
            setattr(self, attr, eval(attr))

    def _get_pf(self) -> bool:
        return self._pf

    def _set_pf(self, value: bool) -> None:
        self._pf = value

    def _get_pma(self) -> bool:
        return self._pma

    def _set_pma(self, value: bool) -> None:
        self._pma = value

    def _get_pmb(self) -> bool:
        return self._pmb

    def _set_pmb(self, value: bool) -> None:
        self._pmb = value

    def _get_pmc(self) -> bool:
        return self._pmc

    def _set_pmc(self, value: bool) -> None:
        self._pmc = value

    def _get_pd(self) -> bool:
        return self._pd

    def _set_pd(self, value: bool) -> None:
        self._pd = value

    def _get_dim(self) -> int:
        return self._dim

    def _set_dim(self, value: int) -> None:
        if value in [1, 2, 3]:
            self._dim = value
        else:
            raise AttributeError(
                f"Dimension of {self.name} cannot be set to {value}. Allowed are 1, 2 or 3."
            )

    def _get_wf_pf(self) -> float:
        return self._wf_pf

    def _set_wf_pf(self, value: float) -> None:
        self._wf_pf = value

    def _get_wf_pma(self) -> float:
        return self._wf_pma

    def _set_wf_pma(self, value: float) -> None:
        self._wf_pma = value

    def _get_wf_pmb(self) -> float:
        return self._wf_pmb

    def _set_wf_pmb(self, value: float) -> None:
        self._wf_pmb = value

    def _get_wf_pmc(self) -> float:
        return self._wf_pmc

    def _set_wf_pmc(self, value: float) -> None:
        self._wf_pmc = value

    def _get_wf_pd(self) -> float:
        return self._wf_pd

    def _set_wf_pd(self, value: float) -> None:
        self._wf_pd = value

    def _get_cfps(self) -> bool:
        return self._cfps

    def _set_cfps(self, value: bool) -> None:
        self._cfps = value

    def _get_s(self) -> float:
        return self._s

    def _set_s(self, value: float) -> None:
        if value < 0.0 or value >= 1.0:
            raise AttributeError(f"Parameter s cannot be set to {value}. Allowed range: [0, 1).")
        self._s = value

    def _get_m(self) -> float:
        return self._M

    def _set_m(self, value: float) -> None:
        if value < 0.0 or value >= 1.0:
            raise AttributeError(f"Parameter M cannot be set to {value}. Allowed range: [0, 1).")
        self._M = value

    def _get_interface_weakening_factor(self) -> float:
        return self._interface_weakening_factor

    def _set_interface_weakening_factor(self, value: float) -> None:
        self._interface_weakening_factor = value

    def _get_force_global_constants(self) -> bool:
        return self._force_global_constants

    def _set_force_global_constants(self, value: bool) -> None:
        self._force_global_constants = value

    def _get_p21_neg(self) -> float:
        return self._p21_neg

    def _set_p21_neg(self, value: float) -> None:
        self._p21_neg = value

    def _get_p21_pos(self) -> float:
        return self._p21_pos

    def _set_p21_pos(self, value: float) -> None:
        self._p21_pos = value

    def _get_p22_neg(self) -> float:
        return self._p22_neg

    def _set_p22_neg(self, value: float) -> None:
        self._p22_neg = value

    def _get_p22_pos(self) -> float:
        return self._p22_pos

    def _set_p22_pos(self, value: float) -> None:
        self._p22_pos = value

    pf = property(_get_pf, _set_pf, doc=_DOC_PF)
    pma = property(
        _get_pma,
        _set_pma,
        doc=_DOC_PMA,
    )
    pmb = property(
        _get_pmb,
        _set_pmb,
        doc=_DOC_PMB,
    )
    pmc = property(_get_pmc, _set_pmc, doc=_DOC_PMC)
    pd = property(
        _get_pd,
        _set_pd,
        doc=_DOC_PD,
    )
    dim = property(
        _get_dim,
        _set_dim,
        doc=_DOC_DIM,
    )

    wf_pf = property(_get_wf_pf, _set_wf_pf, doc=_DOC_WF_PF)
    wf_pma = property(_get_wf_pma, _set_wf_pma, doc=_DOC_WF_PMA)
    wf_pmb = property(_get_wf_pmb, _set_wf_pmb, doc=_DOC_WF_PMB)
    wf_pmc = property(_get_wf_pmc, _set_wf_pmc, doc=_DOC_WF_PMC)
    wf_pd = property(_get_wf_pd, _set_wf_pd, doc=_DOC_WF_PD)
    cfps = property(
        _get_cfps,
        _set_cfps,
        doc=_DOC_CFPS,
    )

    s = property(
        _get_s,
        _set_s,
        doc=_DOC_S,
    )
    M = property(
        _get_m,
        _set_m,
        doc=_DOC_M,
    )

    interface_weakening_factor = property(
        _get_interface_weakening_factor,
        _set_interface_weakening_factor,
        doc=_DOC_INTERFACE_WEAKENING_FACTOR,
    )

    force_global_constants = property(
        _get_force_global_constants,
        _set_force_global_constants,
        doc=_DOC_FORCE_GLOBAL_CONSTANTS,
    )

    p21_neg = property(
        _get_p21_neg,
        _set_p21_neg,
        doc=_DOC_P21_NEG,
    )
    p21_pos = property(_get_p21_pos, _set_p21_pos, doc=_DOC_P21_POS)
    p22_neg = property(
        _get_p22_neg,
        _set_p22_neg,
        doc=_DOC_P22_NEG,
    )
    p22_pos = property(
        _get_p22_pos,
        _set_p22_pos,
        doc=_DOC_P22_POS,
    )


ATTRS_PUCK = inspect.signature(PuckCriterion).parameters.keys()



"""Shear Crimping Failure Criterion."""
import inspect

from ._failure_criterion_base import FailureCriterionBase

_DOC_WF = "Weighting factor of the failure mode (wb or wt)."
_DOC_KC = (
    "Weighting factor of the core material for evaluation of the "
    "maximum allowable load. The default is ``1``."
)
_DOC_KF = (
    "Weighting factor of the face sheets for evaluation of the maximum allowable load. "
    "The default is ``0`` so the face sheet does not contribute to the allowable load. "
    "This property is valid for thin face sheets."
)


class ShearCrimpingCriterion(FailureCriterionBase):
    """Shear Crimping Criterion."""

    __doc__ = f"""Defines the shear crimping failure criterion for sandwich structures.

    Parameters
    ----------
    kc:
        {_DOC_KC}
    kf:
        {_DOC_KF}
    wf:
        {_DOC_WF}
    """

    def __init__(self, *, kc: float = 1.0, kf: float = 0.0, wf: float = 1.0):
        """Create a shear crimping failure criterion for sandwich structures.

        A laminate is classified as sandwich if it has at least one core material.
        A honeycomb is an example.
        """
        super().__init__(name="Shear Crimping", active=True)

        for attr in ATTRS_SHEAR_CRIMPING:
            setattr(self, attr, eval(attr))

    def _get_kc(self) -> float:
        return self._kc

    def _set_kc(self, value: float) -> None:
        self._kc = value

    def _get_kf(self) -> float:
        return self._kf

    def _set_kf(self, value: float) -> None:
        self._kf = value

    def _get_wf(self) -> float:
        return self._wf

    def _set_wf(self, value: float) -> None:
        self._wf = value

    wf = property(_get_wf, _set_wf, doc=_DOC_WF)
    kc = property(_get_kc, _set_kc, doc=_DOC_KC)
    kf = property(_get_kf, _set_kf, doc=_DOC_KF)


ATTRS_SHEAR_CRIMPING = inspect.signature(ShearCrimpingCriterion).parameters.keys()



"""Helper to load composites plugin."""
import os
import pathlib
from typing import Optional

import ansys.dpf.core as dpf
from ansys.dpf.core.misc import get_ansys_path
from ansys.dpf.core.server_types import BaseServer
from ansys.dpf.gate.errors import DPFServerException

from ._versions import version_older_than

# The automatic load of the plug-in can be disabled by the user and so
# all plugins which are required for dpf composites are loaded
_PLUGINS = (
    "Ans.Dpf.Native",
    "mapdlOperatorsCore",
    "Ans.Dpf.FEMutils",
    "Ans.Dpf.EngineeringData",
    "composite_operators",
)


def _load_plugins(
    server: BaseServer,
    awp_root_docker: str,
    ansys_path: Optional[str] = None,
) -> None:
    def get_lib_from_name(name: str) -> str:
        if server.os == "posix":
            return f"lib{name}.so"
        elif server.os == "nt":
            return f"{name}.dll"
        else:
            raise RuntimeError(f"Invalid server os: {server.os}")

    location_in_installer = {
        "composite_operators": {
            "nt": ["dpf", "plugins", "dpf_composites"],
            "posix": ["dpf", "plugins", "dpf_composites"],
        },
    }

    if not ansys_path:
        if server.ansys_path:
            ansys_path = server.ansys_path
        elif server.os == "posix":
            ansys_path = awp_root_docker
        else:
            ansys_path = get_ansys_path()

    if not ansys_path:
        raise RuntimeError("Ansys path is not available. DPF Composites plugin cannot be loaded.")

    for name in _PLUGINS:
        if name in location_in_installer:
            relative_location = location_in_installer[name][server.os]
            library = os.path.join(ansys_path, *relative_location, get_lib_from_name(name))
            if server.os == "posix":
                library = pathlib.Path(library).as_posix()
        else:
            library = get_lib_from_name(name)
        dpf.load_library(library, name, server=server)


def load_composites_plugin(server: BaseServer, ansys_path: Optional[str] = None) -> None:
    r"""Load composites plugins and its dependencies.

    Parameters
    ----------
    server:
    ansys_path:
        Ansys root path, for example C:\\Program Files\\ANSYS Inc\\v232.
        If None, it is assumed that all the plugins and their dependencies
        are found in the PATH/LD_LIBRARY_PATH. If ansys_path
        is set, the composite_operators and
        Ans.Dpf.EngineeringData plugins are loaded from their location
        in the installer.
    """
    if version_older_than(server, "7.0"):
        # March 2023 R.Roos
        # the dpf load_plugin method requires an absolute path because the dpf_composites
        # plugin is not in the search path. The logic below should support all cases
        # except a local gRPC sever on linux (local in-process server on Linux is supported)
        # DPF core team thinks about to expose ANSYS_ROOT_FOLDER so that the absolute path
        # can be easily constructed here.
        # Different versions have different awp roots. Here we just try to load the plugins
        # from different roots.
        possible_awp_roots = [
            "/ansys_inc/ansys/dpf/server_2023_2_pre1",
            "/ansys_inc/ansys/dpf/server_2023_2_pre2",
        ]

        while True:
            awp_root_docker = possible_awp_roots.pop()
            try:
                _load_plugins(server, ansys_path=ansys_path, awp_root_docker=awp_root_docker)
                return
            except DPFServerException:
                if len(possible_awp_roots) == 0:
                    raise
    else:
        for name in _PLUGINS:
            # Since version 7.0, the dpf load_library method can load the plugins
            # by passing their name (as defined in the _PLUGIN variable) in place
            # of the filename.
            # Note greschd July '23: AFAICT, The 'name' argument needs to be unique
            # for each plugin. Otherwise, the plugin may not be loaded if another
            # plugin with the same name is already loaded.
            dpf.load_library(name, name=name, server=server)



"""Functions to get elementary indices based on filter input."""
from typing import Collection, Optional

import numpy as np
from numpy.typing import NDArray

from .constants import Spot
from .layup_info import AnalysisPlyInfoProvider, ElementInfo

__all__ = (
    "get_selected_indices",
    "get_selected_indices_by_dpf_material_ids",
    "get_selected_indices_by_analysis_ply",
)


def _get_rst_spot_index(spot: Spot) -> int:
    """Return the spot in the index in the RST file.

    The order in the RST file is always bottom, top, and middle.
    """
    return [-1, 0, 2, 1][spot]


def get_selected_indices(
    element_info: ElementInfo,
    layers: Optional[Collection[int]] = None,
    nodes: Optional[Collection[int]] = None,
    spots: Optional[Collection[Spot]] = None,
    disable_checks: bool = False,
) -> NDArray[np.int64]:
    """Get elementary indices based on element information, layers, nodes, and spots.

    Parameters
    ----------
    element_info
        Lay-up information for the element.
    layers
        List of selected layers.
    nodes
        List of selected corner nodes.
    spots
        List of selected spots.
    disable_checks:
        Whether to disable checks. Set to ``True`` to disable checks.
        Disabling checks results in better performance but potentially
        cryptic error messages or invalid indices.

    Returns
    -------
    NDArray:
        Array of selected indices.


    Notes
    -----
    Returns an empty selection if any of the collections is empty.
    """
    if layers is None:
        layer_indices: Collection[int] = range(element_info.n_layers)
    else:
        if len(layers) == 0:
            return np.array([])
        layer_indices = layers

    if nodes is None:
        node_indices: Collection[int] = range(element_info.number_of_nodes_per_spot_plane)
    else:
        if len(nodes) == 0:
            return np.array([])
        node_indices = nodes
    if spots is None:
        spot_indices: Collection[int] = range(element_info.n_spots)
    else:
        if len(spots) == 0:
            return np.array([])
        spot_indices = [_get_rst_spot_index(spot) for spot in spots]

    if not disable_checks:
        if not element_info.is_layered:
            raise RuntimeError("Computation of indices is not supported for non-layered elements.")
        if element_info.n_spots == 0:
            raise RuntimeError(
                "Computation of indices is not supported for elements with no spots. This could"
                "mean this is a output that has only been written at the bottom and "
                f"the top of the stack of layers. Element Info: {element_info}."
            )
        if max(layer_indices) >= element_info.n_layers:
            raise RuntimeError(
                f"Layer index {max(layer_indices)} is greater or "
                f"equal to the number of layers: {element_info.n_layers}. "
                f"Element Info: {element_info}."
            )

        if max(node_indices) >= element_info.number_of_nodes_per_spot_plane:
            raise RuntimeError(
                f"Corner node index {max(node_indices)} is greater or "
                f"equal to the number of corner nodes: "
                f"{element_info.number_of_nodes_per_spot_plane}. "
                f"Element Info: {element_info}."
            )

        if max(spot_indices) >= element_info.n_spots:
            raise RuntimeError(
                f"Spot index {max(spot_indices)} is greater or "
                f"equal to the number of spots: {element_info.n_spots}. "
                f"Element Info: {element_info}."
            )

    all_indices = np.zeros(
        len(spot_indices) * len(node_indices) * len(layer_indices), dtype=np.int64
    )
    # Todo: Use numpy. Probably use ravel_multi_index method.
    current_index = 0
    for layer_index in layer_indices:
        layer_start_index = layer_index * element_info.n_corner_nodes * element_info.n_spots
        for spot_index in spot_indices:
            spot_start_index = layer_start_index + spot_index * element_info.n_corner_nodes
            for corner_index in node_indices:
                all_indices[current_index] = spot_start_index + corner_index
                current_index = current_index + 1

    return all_indices


def get_selected_indices_by_dpf_material_ids(
    element_info: ElementInfo, dpf_material_ids: Collection[np.int64]
) -> NDArray[np.int64]:
    """Get selected indices by DPF material IDs.

    This method selects all indices that are in a layer with one of the selected materials.

    Parameters
    ----------
    element_info: ElementInfo
        Lay-up information for the element.
    dpf_material_ids
        Collection of DPF materials.

    Returns
    -------
    NDArray[int64]:
        Selected elementary indices

    """
    layer_indices = [
        index
        for index, mat_id in enumerate(element_info.dpf_material_ids)
        if mat_id in dpf_material_ids
    ]
    return get_selected_indices(element_info, layers=layer_indices)


def get_selected_indices_by_analysis_ply(
    analysis_ply_info_provider: AnalysisPlyInfoProvider, element_info: ElementInfo
) -> NDArray[np.int64]:
    """Get selected indices by analysis ply.

    Selects all indices that are in a layer with the given analysis ply

    Parameters
    ----------
    analysis_ply_info_provider: AnalysisPlyInfoProvider
        Provider for the analysis ply information.
    element_info: ElementInfo
        Lay-up information for the element.

    Returns
    -------
    NDArray[int64]:
        Selected elementary indices.
    """
    layer_index = analysis_ply_info_provider.get_layer_index_by_element_id(element_info.id)
    if layer_index is None:
        raise RuntimeError(
            f"Analysis Ply '{analysis_ply_info_provider.name}' "
            f"is not part of element {element_info.id}"
        )
    else:
        return get_selected_indices(element_info, layers=[int(layer_index)])
